package eve.io.block;
import java.io.StreamCorruptedException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Vector;

import eve.sys.Handle;
import eve.sys.Reflection;
import eve.sys.TimedOutException;
import eve.sys.Wrapper;
import eve.util.ArraySection;
import eve.util.ByteArray;
import eve.util.ByteEncodable;
import eve.util.ByteEncoder;
import eve.util.mVector;
/**
* A RemoteCall with a RemoteCallHandler can be used to invoke methods on an object
* running in one ewe application by a different remote application. The remote call
* handles marshalling the data into text streams which are then sent 
* and reassembled into local data and invoked on the target object. A result
* is sent back indicating success or failure along with any return value. This is very
* useful when synchronizing ewe applications since you do not need to come up with any
* protocols or message passing system. All you need to do is invoke methods remotely.
* <p>
* See the EweSyncDemo application for an example of how to do this.
* <p>
* Please note the following restrictions when using remote calls.
* <ul>
* <li>All individual Object parameters or objects within arrays or vectors must be text
* encodable/decodable in some way.
* </li>
* </ul>
*
* <p>A RemoteCall represents a single method which is specified in its constructor.
* Once you have a RemoteCall object you set parameters for a call by calling the
* add() methods in sequence. The RemoteCall knows the exact type of each parameter
* so the data specified in each add() method is cast to the correct parameter type.
* Each add() method returns the RemoteCall object itself so it is easy to concatenate
* add() calls together.
**/

//##################################################################
public class RemoteCall extends Handle implements ByteEncodable{
//##################################################################

public String method;
public String retType = "";
public String id;
Vector parameters = new Vector();

private static ByteEncoder encoder;
private static synchronized ByteEncoder getByteEncoder()
{
	if (encoder == null) encoder = new ByteEncoder();
	return encoder;
}

char [] types;
//String [] actual;
//String [] values;

/**
* This is used if you intend to send/receive parameters that are classes that are loaded using
* a non-system class loader.
* @deprecated - this is not needed.
**/
public Class classCreator;
/**
* This will hold any exception generated by the remote object when the
* method was invoked. If the exception generated on the remote side
* could not be recreated on this side - it will be set to RemoteCall.UnknownException
and the message will be set to be the same.
**/
public Throwable invocationError;
/**
* This is the timeout for the call. By default it is zero, which will then let the remote call handler
* being used to send the call use its own timeout.
**/
public int timeOut = 0;

/**
* This is initially null but can be set before calling call.
**/
public RemoteCallHandler myHandler;

//##################################################################
public static class UnknownException extends Exception{
//##################################################################

public UnknownException() {super();}
public UnknownException(String message) {super(message);}
//===================================================================
//public UnknownException(Throwable cause) {super(cause);}
//===================================================================
//public UnknownException(String message,Throwable cause) {super(message,cause);}
//===================================================================

//##################################################################
}
//##################################################################

private int curPar = 0;
/**
* Create a new RemoteCall targeted at the specified method. Parameter
* and return specs MUST be provided e.g.<br><br>
* new RemoteCall("aMethod(IILjava/lang/String;)Z");<p>
* Standard Java type specifiers are used, specifically.
* <p>
* B - byte, C - char, S - short, I - int, J - long, F - float, D - Double
* Z - boolean, V - void (for method types only)<br>
* [<i>type-specifier</i> - for an array.<br>
* L<i>className</i>; - for object (NOTE the ending semi-colon).<p> 
* For object (L) you should use / instead of . as the package separator.
* <p>
* The format for a method is: <i>method_name</i>(<i>concatenated parameter type specifiers</i>)<i>return_type</i>
* <p>
* There are no separators between the parameter type specifiers. If the the method is void you should
* use V as the return type.
**/
//===================================================================
public RemoteCall(String method)
//===================================================================
{
	setMethod(method);
}
public RemoteCall(Method method)
{
	setMethod(Reflection.getEncodedMethod(method));
}
public RemoteCall(String methodName, Class[] parameterTypes, Class returnType)
{
	setMethod(Reflection.getEncodedMethod(methodName,parameterTypes,returnType));
}
/**
* This is used to create a RemoteCall from a sequence of bytes. This is used
* by the handler to reconstruct the call after it has been received.
**/
//===================================================================
public RemoteCall(byte[] source,int offset,int length)
throws StreamCorruptedException, ClassNotFoundException
//===================================================================
{
	decodeBytes(source,offset,length);
}
/**
 * Get a copy of a RemoteCall that you can use for another invocation. RemoteCalls should
	only be used once to invoke on the remote object - so if you want to invoke the call
	repeatedly you will need to call this method. However after doing so you will always need
	to add the parameters again.
 * @return A copy of the RemoteCall with no parameters added yet.
 */
//===================================================================
public RemoteCall getCopy()
//===================================================================
{
	return new RemoteCall(method);
}
/**
 * This is used to set a special class creator. This is in case you are going to be
	sending/receiving objects that were loaded from a non-system class loader. The creator
	supplied must have been loaded from the same class loader.
 * @param creator A Class object that was loaded from the same class loader.
 * @return this RemoteCall
 * @deprecated, this is no longer needed - ClassLoaders are cached by mClassLoader
 * and used by the ByteEncoder to create them as needed.
 */
//===================================================================
public RemoteCall setClassCreator(Class creator)
//===================================================================
{
	classCreator = creator;
	return this;
}
//-------------------------------------------------------------------
private void setMethod(String method)
//-------------------------------------------------------------------
{
	this.method = method;
	types = Reflection.getParameterTypesAsEncodedType(method);
	int idx = method.indexOf(')');
	if (idx != -1) retType = method.substring(idx+1);
	//Vm.debug("Create  : "+method);
	//actual = new String[types.length];
	//values = new String[types.length];
}

/*
protected void finalize()
{
	Vm.debug("Finalize: "+method);
}
*/
//-------------------------------------------------------------------
private char getCurType()
//-------------------------------------------------------------------
{
	try{
		return types[parameters.size()];
	}catch(IndexOutOfBoundsException e){
		throw new IllegalStateException("Too many parameters.");
	}
}

/**
 * This is used to add a parameter.
 * @param w a Wrapper with the parameter value. It must be of the correct type.
 * @return this RemoteCall object.
 */
//===================================================================
public RemoteCall addParameter(Wrapper w)
//===================================================================
{
	char type = getCurType();
	if (w.getType() != getCurType()){
		if (w.getType() != Wrapper.OBJECT || getCurType() != '[')
			throw new IllegalArgumentException("Incompatible parameter");
	}
	parameters.add(w);
	return this;
}
/**
* Use this to add a char parameter.
* @param value The character parameter
* @return itself.
*/
//===================================================================
public RemoteCall add(char value)
//===================================================================
{
	return add((long)value);
}
/**
* Use this to add a parameter of type byte, short, int and long.
* @param value The long value to add.
* @return itself.
*/
//===================================================================
public RemoteCall add(long value)
//===================================================================
{
	Wrapper w = new Wrapper();
		switch(getCurType()){
			case 'B' : w.setByte((byte)value); break;
			case 'S' : w.setShort((short)value); break;
			case 'C' : w.setChar((char)value); break;
			case 'I' : w.setInt((int)value); break;
			case 'J' : w.setLong(value); break;
			case 'Z' : w.setBoolean(value != 0); break;
			default:
				throw new IllegalArgumentException("Incompatible parameter");
		}
		return addParameter(w);
}
/**
* Use this to add a parameter of type float or double.
* @param value The double or float value to add.
* @return itself.
*/
//===================================================================
public RemoteCall add(double value)
//===================================================================
{
	Wrapper w = new Wrapper();
		switch(getCurType()){
			case 'D' : w.setDouble(value); break;
			case 'F' : w.setFloat((float)value); break;
			default:
				throw new IllegalArgumentException("Incompatible parameter");
		}
		return addParameter(w);
}
/**
* Use this to add a parameter of type boolean.
* @param value The boolean value to add.
* @return itself.
*/
//===================================================================
public RemoteCall add(boolean value)
//===================================================================
{
	return add(value ? 1 : 0);
}
/**
* Use this to add an object or array parameter (including null) - 
* do not use this to add a Java Wrapper Object for a primitive parameter.. 
* @param value The object or array value to add.
* @return itself.
*/
//===================================================================
public RemoteCall add(Object value) 
//===================================================================
{
	Wrapper w = new Wrapper();
		switch(getCurType()){
			case 'L' : w.setObject(value); break;
			case '[' : 
				if (value != null && !value.getClass().isArray())
					throw new IllegalArgumentException("Incompatible parameter");
				w.setObject(value);
				break;
			default:
				throw new IllegalArgumentException("Incompatible parameter");
		}
		return addParameter(w);
}
/**
 * Add a standard Java parameter. This could be the actual object or
 * a standard Java Wrapper.
 * @param value the parameter to add.
 * @return this RemoteCall.
 */
public RemoteCall addParameter(Object value)
{
	Wrapper w = new Wrapper();
	switch(getCurType()){
		case 'L' : w.setObject(value); break;
		case '[' : 
			if (value != null && !value.getClass().isArray())
				throw new IllegalArgumentException("Incompatible parameter");
			w.setObject(value);
			break;
		default:
			w = Wrapper.toEveWrapper(value,null);
			break;
	}
	return addParameter(w);
}
//===================================================================
public RemoteCall set(Wrapper[] parameters)
//===================================================================
{
	for (int i = 0; i<parameters.length; i++)
		addParameter(parameters[i]);
	return this;
}
/**
 * Use this to set the parameters as standard Java Wrappers - the same
 * as you would use with java.lang.reflect.Method invokation.
 * @param parameters the parameter list.
 * @return this RemoteCall.
 */
public RemoteCall setParameters(Object[] parameters)
{
	if (parameters == null) return this;
	for (int i = 0; i<parameters.length; i++)
		addParameter(parameters[i]);
	return this;
}
//===================================================================
public int encodeBytes(ByteArray dest) throws IllegalStateException
//===================================================================
{
	if (parameters.size() != types.length) throw new IllegalStateException("All the parameters have not been set.");
	ByteArray ba = new ByteArray();
	ByteEncoder bo = getByteEncoder();
	int total = 0;
	synchronized(bo){
		total = bo.encodeField(dest,"<m>",method);
		for (int i = 0; i<parameters.size(); i++)
			bo.encode(ba,(Wrapper)parameters.get(i));
		total += bo.encodeField(dest,"<p>",ba.data,0,ba.length);
	}
	return total;
}
//-------------------------------------------------------------------
private int decodeBytes(byte[] source,int offset,int length)
throws StreamCorruptedException, ClassNotFoundException
//-------------------------------------------------------------------
{
	if (source == null) return length;
	ByteEncoder bo = getByteEncoder();
	synchronized(bo){
		method = bo.decodeStringField(source,offset,length,"<m>");
		ArraySection r = bo.getFieldByteArray(source,offset,length,"<p>",null);
		Wrapper[] all = bo.decode(source,r.startIndex,r.length);
		parameters = mVector.newVector(all);
	}
	return length;
}
/**
* This encodes the call ready for dispatch. You would not need to use this unless you
* were writing a RemoteCallHandler.
* @param dest The destination TextEncoder or null if you want the method to create and return a new one.
* @return The TextEncoder parameter or a new one if it is null.
*/
/*
//===================================================================
public TextEncoder encode(TextEncoder dest) throws IllegalStateException
//===================================================================
{
	if (curPar != types.length) throw new IllegalStateException("The parameters have not been set.");
	if (dest == null) dest = new TextEncoder();
	dest.addValue("<method>",method);
	for (int i = 0; i<types.length; i++)
		dest.addValue(actual[i] == null ? "" : actual[i], values[i] == null ? "" : values[i]);
	return dest;
}
*/


/**
* Invoke the remote call on a target object of the specified class.
* You would not need to use this unless you were writing a RemoteCallHandler.
* @param dest The destination Object to invoke the method on - can be 
* @param r The Reflect object for the destination Object.
* @return An encoded return value.
*/
/*
//===================================================================
public TextEncoder invokeOn(Object dest,Reflect r)
//===================================================================
{
	TextEncoder te = new TextEncoder();
	if (!invokeOn(dest,r,te))
		te.addValue("<error>","Error executing method.");
	return te;
}
*/
//===================================================================
public ByteArray invokeOn(Object dest,Class r)
//===================================================================
{
	return invokeOn(dest,r,null);
}
//===================================================================
public ByteArray invokeOn(Object dest,Class r,RemoteCallTask rct)
//===================================================================
{
	ByteArray ba = new ByteArray();
	if (!invokeOn(dest,r,ba,rct)){
		ByteEncoder bo = getByteEncoder();
		synchronized(bo){
			bo.encodeField(ba,"<err>","Error");
		}
	}
	return ba;
}
static void encodeException(ByteEncoder bo,Throwable t,ByteArray dest,boolean putErrorMarker)
{
	if (putErrorMarker) bo.encodeField(dest,"<err>","Error");
	bo.encodeField(dest,"<ex>",t.getClass().getName());
	String message = t.getMessage();
	if (message != null) bo.encodeField(dest,"<em>",message);
}
static void encodeException(Throwable t,ByteArray dest,boolean putErrorMarker)
{
	ByteEncoder bo = getByteEncoder();
	synchronized(bo){
		encodeException(bo,t,dest,putErrorMarker);
	}
}
//-------------------------------------------------------------------
private boolean returnException(Throwable t,ByteArray dest)
//-------------------------------------------------------------------
{
	//t.printStackTrace();
	encodeException(t,dest,false);
	return false;
}
//
private static Wrapper toProxy(Wrapper w, RemoteCallTask rct) throws ClassNotFoundException
{
	if (rct == null) return w;
	return rct.toProxy(w);
}
private static Wrapper toServer(Class returnClass, Wrapper w, RemoteCallTask rct)
{
	if (rct == null) return w;
	return rct.toOutgoingRemoteInterface(returnClass, w);
}
//-------------------------------------------------------------------
protected boolean invokeOn(Object dest,Class r,ByteArray ba, RemoteCallTask rct)
//-------------------------------------------------------------------
{
	if (ba == null) ba = new ByteArray();
	if (r == null) r = dest == null ? null : dest.getClass();
	if (r == null) return false;
	int idx = method.indexOf('(');
	if (idx == -1) return false;
	Method md = Reflection.getMethod(r,method,false);
	//System.out.println("Invoke: "+md);
	if (md == null) 
		return returnException(new NoSuchMethodError(method),ba);
	Wrapper [] p = new Wrapper[parameters.size()];
	try{
		for (int i = 0; i<p.length; i++)
			p[i] = toProxy((Wrapper)parameters.get(i),rct);
		Wrapper got = new Wrapper();
		Reflection.invoke(dest,md,p,got);
 		Class rt = md.getReturnType();
 		boolean returnsValue = (rt != null && !rt.equals(Void.TYPE)); 
 		ByteEncoder bo = getByteEncoder();
 		synchronized(bo){
	 		bo.encodeField(ba,"<OK>","");
	 		if (returnsValue) try{
	 			got = toServer(rt,got,rct);
	 			bo.encodeField(ba,"<ret>",got);
	 		}catch(Exception e){
	 			//e.printStackTrace();
	 		}
 		}
		return true;
	}catch(InvocationTargetException te){
		return returnException(te.getTargetException(),ba);
	}catch(Throwable e){
		return returnException(e,ba);
	}
}
/**
* This is used to tell the RemoteCall that it is in the process of transmission
* to the remote object.
* You would not need to use this unless you were writing a RemoteCallHandler.
**/
//===================================================================
public void sent()
//===================================================================
{
	set(Running);
}
/**
 * This sets the timeout time for this RemoteCall. This overrides the timeout of the
 * remote call handler being used to send the call.
 * @param milliseconds the timeout in milliseconds. Setting it to zero will cause the
 * remote call handler timeout to be used instead.
 * @return this RemoteCall.
 */
//===================================================================
public RemoteCall setTimeout(int milliseconds)
//===================================================================
{
	timeOut = milliseconds;
	return this;
}
/**
* This is set true once a reply has been received.
**/
//===================================================================
public boolean replied = false;
//===================================================================

//-------------------------------------------------------------------
private static String get(ByteArray ba,String name)
//-------------------------------------------------------------------
{
	ByteEncoder bo = getByteEncoder();
	synchronized(bo) {
		try{
			return (String)bo.decodeObjectField(ba.data,0,ba.length,name,null);
		}catch(Exception e){
			return null;
		}
	}
}

static Throwable getException(ByteArray reply)
{
	String ie = get(reply,"<err>");
	if (ie == null) return null;
	Throwable invocationError = null;
	String msg = null;
	String ex = get(reply,"<ex>");
	if (ex != null){
		msg = get(reply,"<em>");
		Class r = Reflection.forName(ex);
		if (r != null){
			Constructor c = Reflection.getConstructor(r,msg == null ? "" : "Ljava/lang/String;");
			if (c == null) invocationError = new UnknownException(ex+(msg == null ? "" : (" - "+msg)));
			else{
				Wrapper [] inv = new Wrapper[msg == null ? 0 : 1];
				if (msg != null) inv[0] = new Wrapper().setObject(msg);
				try{
					invocationError = (Throwable)Reflection.newInstance(c,inv);
				}catch(Exception e){
					invocationError = null;
				}
				if (invocationError == null) invocationError = new UnknownException(ex+(msg == null ? "" : (" - "+msg)));
			}
		}
	}else{
		if (msg == null) msg = "";
		invocationError = new UnknownException(msg);
	}
	return invocationError;
}
/**
* This is used to provide a reply to the remote call from an encoded
* reply from the remote host. It will then prepare
* the return value Wrapper and set its flags to be successfull.
* You would not need to use this unless you were writing a RemoteCallHandler.
**/
//===================================================================
public synchronized void reply(ByteArray reply)
//===================================================================
{
	if (error != null) return;
	try{
		String ie = get(reply,"<err>");
		if (ie == null){
			try{
				ByteEncoder bo = getByteEncoder();
				synchronized(bo){
					returnValue = bo.decodeField(reply.data,0,reply.length,"<ret>",null);
					if (returnValue == null)
						returnValue = new Wrapper().setType(Wrapper.VOID);
				}
				set(Succeeded);
			}catch(Exception e){}
		}else{
			//fail(new RemoteCallException(this,getException(reply),null));
			fail(new InvocationTargetException(getException(reply)));
			/*
			String ex = get(reply,"<ex>");
			if (ex != null){
				String msg = get(reply,"<em>");
				Class r = Reflection.forName(ex);
				if (r == null){
					invocationError = new UnknownException(ex+(msg == null ? "" : (" - "+msg)));
				}else{
					Constructor c = Reflection.getConstructor(r,msg == null ? "" : "Ljava/lang/String;");
					if (c == null) invocationError = new UnknownException(ex+(msg == null ? "" : (" - "+msg)));
					else{
						Wrapper [] inv = new Wrapper[msg == null ? 0 : 1];
						if (msg != null) inv[0] = new Wrapper().setObject(msg);
						try{
							invocationError = (Throwable)Reflection.newInstance(c,inv);
						}catch(Exception e){
							invocationError = null;
						}
						if (invocationError == null) invocationError = new UnknownException(ex+(msg == null ? "" : (" - "+msg)));
					}
				}
				fail(new RemoteCallException(this,invocationError,null));
			}
			*/
		}
	}finally{
		replied = true;
		setFlags(Stopped,0);
	}
}
/**
* This calls and waits for the remote call to complete, fail or timeout. The RemoteCall
* itself is returned and you should check to see if the "error" member is null.
**/
//===================================================================
public RemoteCall callAndWait(RemoteCallHandler handler,String targetCode)
//===================================================================
{
	handler.call(this,targetCode);
	while(true){
		try{
			waitUntilStopped();
			return this;
		}catch(InterruptedException e){
		}
	}
}

/**
 * Invokes the remote call using the specified handler as the transport mechanism.
 * @param handler the RemoteCallHandler handling the call.
 * @param targetCode the target object.
 * @return a Wrapper containing the return value. Use getInt(), getDouble(), getObject(), etc. on the
	returned value to convert to the appropriate type.
 * @exception RemoteCallException if an error occured during the RemoteCall.
 * @exception TimedOutException if the remote call timed out before a reply was made.
 */
//===================================================================
public Wrapper call(RemoteCallHandler handler,String targetCode) throws InvocationTargetException, RemoteCallException, TimedOutException
//===================================================================
{
	callAndWait(handler,targetCode);
	if ((check() & Success) == 0){
		if (error instanceof TimedOutException) {
			error.fillInStackTrace();
			throw (TimedOutException)error;
		}
		else if (error instanceof RemoteCallException) throw (RemoteCallException)error;
		else if (error instanceof InvocationTargetException) throw (InvocationTargetException)error;
		else throw new RemoteCallException(this,error,"Unknown RemoteCall error");
	}
	return (Wrapper)returnValue;
}
//===================================================================
public RemoteCall callAsync(RemoteCallHandler handler,String targetCode)
//===================================================================
{
	if (handler == null) handler = myHandler;
	myHandler.call(this,targetCode);
	return this;
}
//===================================================================
public RemoteCall callAsync()
//===================================================================
{
	return callAsync(myHandler,null);
}
//===================================================================
public Wrapper getReturnValue()
//===================================================================
{
	return (Wrapper)returnValue;
}
//===================================================================
public Throwable getException()
//===================================================================
{
	return invocationError;
}
/**
 * Invokes the remote call using the myHandler value as the transport mechanism.
 * @param targetCode the target object.
 * @return a Wrapper containing the return value. Use getInt(), getDouble(), getObject(), etc. on the
	returned value to convert to the appropriate type. A Wrapper is always returned, even if
	the method is a void method.
 * @exception RemoteCallException if an error occured during the RemoteCall.
 */
//===================================================================
public Wrapper call(String targetCode) throws RemoteCallException, TimedOutException, InvocationTargetException
//===================================================================
{
	if (myHandler == null) throw new IllegalStateException("RemoteCallHandler not set.");
	return call(myHandler,targetCode);
}
/**
 * Invokes the remote call using the myHandler value as the transport mechanism and specifying a null (default) target object.
 * The method waits until the call returns.
 * @return a Wrapper containing the return value. Use getInt(), getDouble(), getObject(), etc. on the
	returned value to convert to the appropriate type. A Wrapper is always returned, even if
	the method is a void method.
 * @exception RemoteCallException if an error occured during the RemoteCall.
 * @exception TimedOutException if the operation took too long to complete.
 */
//===================================================================
public Wrapper call() throws RemoteCallException, TimedOutException, InvocationTargetException
//===================================================================
{
	if (myHandler == null) throw new IllegalStateException("RemoteCallHandler not set.");
	return call(myHandler,null);
}
/**
* Create a remote call from a received encoded call. You would not use this
* unless you were writing your own RemoteCallHandler.
**/
/*
//===================================================================
public RemoteCall(TextDecoder from)
//===================================================================
{
	setMethod(from.getValue("<method>"));
	int sz = from.size();
	int t = 0;
	for(int i = 0; i<sz; i++){
		String v = from.getName(i);
		if (v != null)
			if (v.length() > 0)
				if (v.charAt(0) == '<')
					continue;
		actual[t] = v;
		values[t] = from.getValue(i);
		t++;
	}
}
*/
//===================================================================
public String toString()
//===================================================================
{
	return method;
}
//##################################################################
}
//##################################################################

