package eve.ui;
/**
 * This Interface is implemented by Controls or other Objects which can be
 * scrolled or tracked in some way. Although it applies usually to Controls
 * it can be applied to more abstract Objects, such as media clips.<p>
 * A ScrollClient is controled by a ScrollServer - the ScrollServer usually
 * has and takes input from the various user scrolling controls (e.g. like 
 * a scroll bar or a track bar) and then translates the user input into scroll
 * commands which are sent to the ScrollClient via the doScroll() method.<p>
 *  
 * @author Michael L Brereton
 *
 */
//##################################################################
public interface ScrollClient extends IScroll {
//##################################################################
	
/**
 * This is called by the ScrolledServer to indicate that the user has requested
 * a scroll.
 * @param which either Horizontal or Vertical
 * @param action one of PageHigher, PageLower, ScrollHigher, ScrollLower, TrackTo
 * @param value this is used for the TrackTo action and indicates the position to
 * track to.
 */
public void doScroll(int which,int action,int value);
/**
 * Notifies the ScrollClient of its new ScrollServer. 
 * @param server the ScrollServer for the ScrollClient.
 */
public void setServer(ScrollServer server);
/**
 * Returns the ScrollServer for the ScrollClient if present.
 */
public ScrollServer getServer();
/**
 * Get the full number of units either vertically or horizontally that the ScrollClient
 * contains. The units are always integer values and their meaning is arbitrary. In a
 * Control a unit may be pixels or it may be rows or columns. In video clip a unit
 * may be an individual frame.
 * @param which either Horizontal or Vertical.
 * @return the number of units for the specified dimension.
 */
public int getActual(int which);
/**
 * Given a display size, this method should return the number of units that can be fully
 * displayed within the display size for the specified dimension.
 * @param which either Horizontal or Vertical.
 * @param forVisibleSize the size (usually in pixels) allowed to be displayed on screen.
 * @return the number of units that can be fully displayed in that size.
 */
public int getVisible(int which,int forVisibleSize);
/**
 * This should return which unit is currently being displayed. If more than one unit can
 * be displayed at a time, it should return the first unit displayed (usually the top
 * unit for vertical displays, and the left unit for horizontal displays).
 * @param which either Horizontal or Vertical.
 * @return the unit currently being displayed.
 */
public int getCurrent(int which);
/**
 * This method should return whether a ScrollBar (or some other form of scrolling
 * user control) is necessary for the specified dimension. Usually this is easily
 * determined as being if getActual(which) is greater than getVisible(which,forVisibleSize)<p>
 * However a ScrollClient can return true always if it always wishes a ScrollBar to
 * be present, or false always if it never wishes a ScrollBar to be present.  
 * @param which either Horizontal or Vertical.
 * @param forVisibleSize the size (usually in pixels) allowed to be displayed on screen.
 * @return whether a ScrollBar is necessary.
 */
public boolean needScrollBar(int which,int forVisibleSize);
/**
 * This method should return whether the ScrollClient can actually scroll in a particular
 * direction, given a starting position in units.
 * @param orientation either Horizontal or Vertical.
 * @param direction either GoHigher or GoLower.
 * @param position a starting position in units.
 * @return true if it can scroll in that direction, false if not.
 */
public boolean canGo(int orientation,int direction,int position);
//##################################################################
}
//##################################################################

