package eve.ui;

import java.util.Vector;

import eve.fx.Color;
import eve.fx.Font;
import eve.fx.FontMetrics;
import eve.fx.Fx;
import eve.fx.Graphics;
import eve.fx.IImage;
import eve.fx.IconAndText;
import eve.fx.Image;
import eve.fx.ImageTool;
import eve.fx.Metrics;
import eve.fx.Rect;
import eve.fx.gui.IEventModifiers;
import eve.fx.gui.IKeys;
import eve.fx.gui.RawEvent;
import eve.fx.gui.SIPConstants;
import eve.fx.gui.SurfaceKeyEvent;
import eve.fx.gui.WindowSurface;
import eve.sys.Event;
import eve.sys.EventDispatcher;
import eve.sys.EventListener;
import eve.sys.Vm;
import eve.ui.event.ControlEvent;
import eve.ui.event.KeyEvent;
import eve.ui.event.MenuEvent;
import eve.ui.event.SoftKeyEvent;
import eve.util.Grid;
import eve.util.mString;
import eve.util.mVector;
/**

A SoftKeyBar is used on Smartphone or other such systems where general purpose
hardware keys are used to control inputs to the software in a way that changes
depending on the situation.<p>

SoftKeyBars provide two types of controls - buttons which simply fire an action when the associated
softkey is pressed, or menus which display a general Menu when pressed.<p>

The SoftKeyBar is not a UI Control since it may actually be implemented using the system's native
UI system - but it is fully responsible for configuring the appearance on screen and handling the
softkey presses. The only thing an application has to do is to trap the events sent out by
the SoftKeyBar or override the event handling of the SoftKeyBar.<p>

Because only one SoftKeyBar can be displayed at a time, you display it by calling Window.setSoftKeyBar()
on the containing Window.
**/
//##################################################################
public class SoftKeyBar implements EventListener{
//##################################################################
/**
You can use this as an identifier for bar.
**/
public String name;

/**
 * Set this to false if you do not want numbers to appear
 * in the Menus within SoftKeys.
 */
public static boolean showMenuNumbers = true;
/**
 * This is the preferred height of the softkey bar.
 * Under windows it defaults to the height of a "standard"
 * Menu item.
 */
public static int preferredBarHeight = 20;

private static String wv = Vm.getProperty("windows.version","win32");
private static boolean isMobile6 = wv.equalsIgnoreCase("mobile 6");
private static boolean isMobile5 = isMobile6||wv.equalsIgnoreCase("mobile 5");
/**
The default background color for the soft keys. Defaults to MediumBlue
**/
public static Color defaultBackgroundColor = !isMobile5 ? new Color(140,215,247) : new Color(0x37,0x91,0xd3);
public static Color defaultTextColor = !isMobile5 ? Color.Black : Color.White;
private static int upColor(int v)
{
	v += 26;
	if (v > 255) v = 255;
	return v;
}
static {
	Object c = Gui.guiValues.get("ActiveTitleBarColor");
	if (c instanceof Color){
		Color cc = (Color)c;
		defaultBackgroundColor = new Color(upColor(cc.getRed()),upColor(cc.getGreen()),upColor(cc.getBlue()));
	}
	c = Gui.guiValues.get(GuiValueKeys.TitleBarTextColor);
	if (c instanceof Color) defaultTextColor = ((Color)c).getCopy();
}
private static Color topStripe = new Color(198,219,239);
private static Color bottomStripe = new Color(8,28,41);
/**
This is false by default and if you set it true, then a hide() will cause
the SoftKeyBar to disappear entirely, instead of just becoming empty - which
is the default.
**/
public static boolean hideRemovesSoftKeyBar = false;
private Vector keys = new Vector();

private EventDispatcher dispatcher;

public int barModifiersToSet = Control.DrawFlat;
public int barModifiersToClear = 0;

public final static SoftKeyBar empty = new SoftKeyBar();

//===================================================================
public SoftKeyBar()
//===================================================================
{
	
}
//===================================================================
public SoftKeyBar(EventListener listener)
//===================================================================
{
	this();
	addListener(listener);
}
//===================================================================
public synchronized void addListener(EventListener el)
//===================================================================
{
	if (dispatcher == null) dispatcher = new EventDispatcher();
	dispatcher.dontUseSeparateThread = true;
	dispatcher.addListener(el,false);
}
//===================================================================
public synchronized void removeListener(EventListener el)
//===================================================================
{
	dispatcher.removeListener(el);
	if (dispatcher.isEmpty()) dispatcher = null;
}

//-------------------------------------------------------------------
protected void sendToListeners(Event ev)
//-------------------------------------------------------------------
{
	if (dispatcher == null) return;
	dispatcher.dispatch(ev);
}
//===================================================================
public void clearKeys()
//===================================================================
{
	keys.clear();
}
/**
This is the default font.
**/
public static Font defaultFont;

private Font font;
/**
 * Set the Font to be used by the bar if possible.
 * @param f the Font to use if this is allowed by the system. If setting of the Font is not allowed,
 * then this method will have no effect.
 */
//===================================================================
public void setFont(Font f)
//===================================================================
{
	font = f;
}
/** This is the value 1 for all Smartphones with two soft keys. **/
//public static int LEFT_KEY = 0x1;
/** This is the value 2, for convenient use on Smartphones with two soft keys. **/
//public static int RIGHT_KEY = 0x2;


/**
 * Returns true if the softkey displays on screen supports the display of images, false if not.
 */
//===================================================================
public static boolean supportsImages()
//===================================================================
{
	return true;
}
static int rawKeys = 2;
/**
Returns the number of keys supported by the system.
**/
//===================================================================
public static int numberOfKeys()
//===================================================================
{
	return isSingleMenuOnly() ? 1 : rawKeys; //FIXME
}
/**
 * Some platforms have only one softkey that always brings up a Menu (like
 * Android). If this is the case then you should never use buttons, you should
 * use menus instead. 
 * @return
 */
public static boolean isSingleMenuOnly()
{
	return ((Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_HIDDEN_SOFTKEYS) != 0);
}
/**
 * A return value for getType() - indicates that only a single MENU softkey
 * is available (e.g. on an Android platform).
 */
public static final int TYPE_MENU = -1;
/**
 * A return value for getType() - indicates that only one softkey exists but
 * it is visible and can be a button or a menu.
 */
public static final int TYPE_SINGLE = 1;
/**
 * A return value for getType() - indicates that two softkeys exists and either
 * can be a button or a menu.
 */
public static final int TYPE_DOUBLE = 2;

/**
 * A return value for getType() - indicates that the system does not use
 * softkeys.
 */
public static final int TYPE_NONE = 0;
/**
 * Determine the type of SoftKeys available.
 * @return one of the TYPE_XXX values.
 */
public static int getType()
{
	if (!Gui.hasSoftKeys) return TYPE_NONE;
	if (isSingleMenuOnly()) return TYPE_MENU;
	return numberOfKeys() == 2 ? TYPE_DOUBLE : TYPE_SINGLE;
}

//===================================================================
public void onEvent(Event ev)
//===================================================================
{
	SoftKeyEvent se = null;
	if (ev instanceof MenuEvent && ev.type == MenuEvent.SELECTED){
		se = new SoftKeyEvent();
		Object got = ((MenuEvent)ev).selectedItem;
		if (got instanceof MenuItem){
			se.selectedItem = ((MenuItem)got);
			se.action = se.selectedItem.action;
			se.proxy = se.selectedItem.data instanceof Control ? (Control)se.selectedItem.data : null;
		}else if (got != null) se.action = got.toString();
	}else if (ev instanceof ControlEvent && ev.type == ControlEvent.PRESSED){
		se = new SoftKeyEvent();
		se.action = ((ControlEvent)ev).action;
		se.proxy = ev.target instanceof Control ? ((Control)ev.target).promptControl : null;
	}
	if (se != null) {
		Window w = barContainer == null ? null : barContainer.getWindow();
		if (w != null) Gui.closePopups(w);
		se.bar = this;
		for (int i = 0; i<keys.size(); i++){
			Button mb = (Button)keys.get(i);
			if (mb == ev.target) {
				se.whichKey = i+1;
				se.target = ev.target;
			}
		}
		sendToListeners(se);
	}	
}
//===================================================================
public FontMetrics getFontMetrics()
//===================================================================
{
	if (font == null) font = defaultFont;
	if (font == null) return Application.mainApp.getFontMetrics();
	else return Application.mainApp.getFontMetrics(font);
}
IImage fixImage(IImage im)
{
	if (im == null || im instanceof IconAndText) return im;
	int h = im.getHeight();
	int w = im.getWidth();
	double scale = 0;
	if (preferredBarHeight <= 26){
		if (h > 16) scale = (double)16/h;
	}else if (preferredBarHeight <= 35){
		if (h < 24 || h > 32) scale = (double)24/h;
	}else
		if (h < 32 || h > (preferredBarHeight*7)/10)
			scale = (double)24/h;
			//scale = (double)32/h;
	if (scale != 0)
		return ImageTool.scale(im,(int)(scale*w),(int)(scale*h),0).toPicture();
	return im;
}

private softkeyEntry [] alreadySet = new softkeyEntry[2];

class softkeyEntry{
	public softkeyEntry(String text, IImage icon, IImage image, Menu menu, Button button)
	{
		this.text = text;
		this.icon = icon;
		this.image = image;
		this.menu = menu;
		this.button = button;
	}
	public void addTo(Menu m)
	{
		if (menu != null){
			int max = menu.itemsSize();
			for (int i = 0; i<max; i++){
				m.addItem(menu.getItemAt(i));
			}
		}else if (button != null){
			m.addItem(createMenuItem(button));
		}else{
			MenuItem mi = new MenuItem();
			int w = text.indexOf('|');
			if (w == -1){
				mi.label = text;
				mi.action = text;
			}else{
				mi.action = text.substring(w+1);
				text = mi.label = text.substring(0,w);
			}
			if (icon != null) {
				mi.image = new IconAndText(icon,text,getFontMetrics());
				((IconAndText)mi.image).textColor = null;
			}else if (image != null){
				mi.image = image;
			}
			m.addItem(mi);
		}
	}
	String text;
	IImage icon;
	IImage image;
	Button button;
	Menu menu;
}

private Button updateKeys()
{
	softkeyEntry left = alreadySet[0];
	softkeyEntry right = alreadySet[1];
	if (left == null && right != null){
		left = right;
		right = null;
	}
	//
	// If only one has been set so far and it is not hidden 
	// then go as normal.
	//
	if (!isSingleMenuOnly() && right == null)
		return _setKey(1,left.text,left.icon,left.image,left.menu,left.button);
	//
	final Menu lm = left == null ? null : left.menu;
	final Menu rm = right == null ? null : right.menu;
	Menu m = new Menu(){
		Menu leftMenu, rightMenu;
		{
			leftMenu = lm;
			rightMenu = rm;
		}
		public boolean prepareForDisplay()
		{
			if (leftMenu != null) leftMenu.prepareForDisplay();
			if (rightMenu != null) rightMenu.prepareForDisplay();
			if (leftMenu != null || rightMenu != null){
				removeAll();
				softkeyEntry left = alreadySet[0];
				softkeyEntry right = alreadySet[1];
				if (left != null)
					left.addTo(this);
				if (right != null) {
					addItem("-");
					right.addTo(this);
				}
			}
			return true;
		}
	};
	m.text = "Menu";
	if (left != null)
		left.addTo(m);
	if (right != null) {
		m.addItem("-");
		right.addTo(m);
	}
	return _setKey(1,"Menu",null,null,m,null);
}
//-------------------------------------------------------------------
private Button setKey(int which, String text, IImage icon, IImage image, Menu menu, Button button) throws IllegalArgumentException
//-------------------------------------------------------------------
{
	//System.out.println("SK: "+which+", "+text+", "+icon+", "+image+", "+menu+", "+button);
	if (which > 2) which = 2;
	else if (which < 1) which = 1;
	if (numberOfKeys() == 2) return _setKey(which,text,icon,image, menu, button);
	//
	// There is only one key.
	//
	alreadySet[which-1] = new softkeyEntry(text,icon,image,menu,button);
	return updateKeys();
}
private Button _setKey(int which, String text, IImage icon, IImage image, Menu menu,Button button) throws IllegalArgumentException
{
	if (text == null) text = "";
	String action = null;
	int w = text.indexOf('|');
	if (w != -1) {
		action = text.substring(w+1);
		text = text.substring(0,w);
	}
	menu = fixMenu(menu);
	Button mb = menu == null ? new Button(text) : new ButtonPullDownMenu(text,menu);
	if (isMobile5){
		if (menu != null) {
			menu.foreGround = Color.DarkBlue;
			menu.font = defaultFont;
		}
	}
	mb.foreGround = defaultTextColor;
	mb.arrowDirection = 0;
	mb.modify(0,mb.MakeMenuAtLeastAsWide);
	if (supportsImages()){
		if (image != null)
			Gui.iconize(mb,fixImage(image),false,getFontMetrics());		
		else if (icon != null)
			Gui.iconize(mb,fixImage(icon),true,getFontMetrics());
	}else if (image != null)
		throw new IllegalArgumentException();
	//	mb.image = new IconAndText(icon,text,getFontMetrics());
	mb.action = action != null ? action : text;
	while(which > keys.size()) keys.add(null);
	keys.setElementAt(mb,which-1);
	mb.addListener(this);
	mb.modify(mb.NoFocus,mb.TakesKeyFocus);
	if (menu != null) mb.menuState.autoSelectFirst = true;
	if (button != null){
		mb.textPosition = button.textPosition;
		mb.promptControl = button;
		mb.action = button.action == null ? mb.text : button.action;
	}
	return mb;
}
/**
 * Create a MenuItem suitable for use in a menu assigned to the SoftKeyBar.
 * @param label the display text for the item.
 * @param action an action String assigned to the item. The handleAction()
 * method will be called on the Form with this value when the item is selected.
 * @param icon an optional icon to be displayed next to the item. 
 * @return the created MenuItem.
 */
//===================================================================
public MenuItem createMenuItem(String label, String action, IImage icon)
//===================================================================
{
	if (action == null) action = label;
	MenuItem mi = new MenuItem(label);
	mi.action = action;
	mi.label = label;
	if (icon != null) {
		mi.image = new IconAndText(icon,label,getFontMetrics());
		((IconAndText)mi.image).textColor = null;
	}
	return mi;
}
/**
Return a MenuItem that will be used as a proxy for a specific buton. When
the menu item is selected via the soft key bar, the specified button will fire 
an action event (ControlEvent.PRESSED)
**/
//===================================================================
public MenuItem createMenuItem(Button button)
//===================================================================
{
	MenuItem mi = createMenuItem(button.text, button.action, button.image);
	mi.data = button;
	return mi;
}
/**
Set one of the SoftKeyBar keys to be a proxy for an existing Button. Pressing that
soft key will cause the specified button to fire of an action event (ControlEvent.PRESSED)
* @param which the key to set, either 1 (left) or 2 (right).
* @param button the button that will be activated by the Soft Key.
* @return this SoftKeyBar
*/
//===================================================================
public SoftKeyBar setKey(int which, Button button)
//===================================================================
{
	Button mb = setKey(which,button.text,button.image,(Image)null,null,button);
	if (button.image instanceof IconAndText){
		((IconAndText)button.image).textColor = null;
		((IconAndText)button.image).fontMetrics = getFontMetrics();
	}
	return this;
}
/**
 * Return the Control the SoftKey is set for. 
 * This is either a Button or a Menu or null. If baseKeyOnly is true,
 * then if the key is a Menu, it will still return the Button that
 * activates the menu. 
 * @param whichKey either 1 or 2.
 * @param baseKeyOnly set true to get the buttons in the SoftKeyBar only, false
 * to get the Menu if either of the Buttons activates a Menu.
 * @return the Control the SoftKey is set for. This is either a Button or a Menu or null.
 */
public Control getKey(int whichKey, boolean baseKeyOnly)
{
	if (whichKey < 1 || whichKey > keys.size()) return null;
	Control c = (Control)keys.get(whichKey-1);
	if (baseKeyOnly || c == null) return c;
	Menu m = c.getMenu();
	if (m != null) return m;
	return c;
	
}
/**
 * Return the Control the SoftKey is set for. This is either a Button or a Menu or null.
 * @param whichKey either 1 or 2.
 * @return the Control the SoftKey is set for. This is either a Button or a Menu or null.
 */
public Control getKey(int whichKey)
{
	return getKey(whichKey,false);
}
/**
Create a MenuItem for a particular Object.
@param obj - this can be a String or an Button, each of which creates a normal MenuItem,
or it can be a Menu or Vector or Object[] or a Container, in which case a MenuItem that has
a sub-menu will be created.
@return the MenuItem created.
*/
//===================================================================
public MenuItem createMenuItemFor(Object obj) throws IllegalArgumentException
//===================================================================
{
	if (obj instanceof String){
		String label = mString.leftOf(obj.toString(),'|');
		String action = mString.rightOf(obj.toString(),'|');
		if (action.length() == 0) action = label;
		return createMenuItem(label,action,(IImage)null);
	}else if (obj instanceof Button){
		return createMenuItem((Button)obj);
	}else if (obj instanceof Menu || obj instanceof Vector || obj instanceof Object[] || obj instanceof Container){
		Menu m = createMenuFor(obj);
		MenuItem mi = new MenuItem();
		mi.subMenu = m;
		mi.label = m.text;
		return mi;
	}else if (obj instanceof MenuItem){
		return (MenuItem)obj;
	}else throw new IllegalArgumentException();
}
/**
Create a Menu for a specific set of Objects, placed in a Vector, or a Container or an Object
array.
Each item in the Vector can be an Button or a String or a MenuItem or another Vector
or Menu.
**/
//===================================================================
public Menu createMenuFor(Object obj)
throws IllegalArgumentException
//===================================================================
{
	Menu m = new Menu();
	Vector v = null;
	if (obj instanceof Object[]) v = mVector.newVector((Object[])obj);
	else if (obj instanceof Vector) v = (Vector)obj;
	else if (obj instanceof Container) v = Form.gatherButtons((Container)obj,null);
	else if (obj instanceof Menu) return fixMenu((Menu)obj);
	else throw new IllegalArgumentException();
	for (int i = 0; i<v.size(); i++)
		m.addItem(createMenuItemFor(v.get(i)));
	return m;
}


/**
 * Set one of the SoftKeys to be a button or a Menu, depending on the type of the obj
 * parameter.
* @param which the key to set, either 1 (left) or 2 (right.
 * @param obj this can be either a String or Button (either of which creates a single
	button for the Soft Key) or a Vector or Object[] or Container or Menu - in which
	case a Menu will be created for the Soft Key.
 * @param label if the obj parameter gets converted to a Menu, then this parameter will
	be the label for the menu.
 * @return this SoftKeyBar
 * @exception IllegalArgumentException 
 */
//===================================================================
public SoftKeyBar setKey(int which, Object obj, String label) throws IllegalArgumentException
//===================================================================
{
	if (obj instanceof String) setKey(which,obj.toString(),(IImage)null,(IImage)null,null,null);
	else if (obj instanceof Button) setKey(which,(Button)obj);
	else if (obj instanceof Menu || obj instanceof Vector || obj instanceof Container || obj instanceof Object[]){
		Menu m = createMenuFor(obj);
		setKey(which,label,(IImage)null,(IImage)null,m,null);
	}
	return this;
}
/**
 * @param which The key to set, starting from 1.
 * @param image The image to use for the key.
 * @param menu An optional menu to display when the key is pressed. If this is null the key
 * will just be treated as a button.
 * @exception IllegalArgumentException if the system does not support displaying images on the
 * soft keys.
 */
//===================================================================
public SoftKeyBar setKey(int which, IImage image, Menu menu) throws IllegalArgumentException
//===================================================================
{
	setKey(which,"",(IImage)null,image,menu,null);
	return this;
}
/**
 * @param which The key to set, starting from 1.
 * @param text The text for the key. If this is in the form "Text_1|Text_2" then "Text_1" is used
 * as the text for the key and "Text_2" is used as the action for the key.
 * @param icon An image to display as an icon next to the text if possible. If this is not possible
 * then the icon is ignored.
 * @param menu An optional menu to display when the key is pressed. If this is null the key
 * will just be treated as a button.
 * @exception IllegalArgumentException if the system does not support displaying images on the
 * soft keys.
 */
//===================================================================
public SoftKeyBar setKey(int which, String text, IImage icon, Menu menu) throws IllegalArgumentException
//===================================================================
{
	setKey(which,text,icon,null,menu,null);
	return this;
}
/**
 * @param which The key to set, starting from 1.
 * @param text The text for the key. If this is in the form "Text_1|Text_2" then "Text_1" is used
 * as the text for the key and "Text_2" is used as the action for the key.
 * @param menu An optional menu to display when the key is pressed. If this is null the key
 * will just be treated as a button with the Text and action for the button being the same.
 * @exception IllegalArgumentException if the system does not support displaying images on the
 * soft keys.
 */
//===================================================================
public SoftKeyBar setKey(int which, String text, Menu menu)
//===================================================================
{
	setKey(which,text,(IImage)null,null,menu,null);
	return this;
}
public static Menu insertAsSubMenu(Menu mainMenu, String asName, Vector menuItems)
{
	MenuItem [] mm = new MenuItem[menuItems.size()];
	menuItems.copyInto(mm);
	Menu m = new Menu(mm,asName);
	mainMenu.addItem(m);
	m.foreGround = mainMenu.foreGround;
	fixMenu(m);
	return m;
}
static Menu splitMenu(String items,Vector dest)
{
	Menu m = dest == null ? new Menu() : null;
	char s = (items.indexOf('|') == -1) ? ',':'|';
	String[] all = mString.split(items,s);
	for (int i = 0; i<all.length;){
		MenuItem mi = new MenuItem(all[i++]);
		if (s == '|') mi.action = all[i++];
		if (dest != null) dest.add(mi);
		else m.addItem(mi);
	}
	if (m != null) fixMenu(m);
	return m;
}
public static Menu insertAsSubMenu(Menu mainMenu, String asName, String items)
{
	MenuItem mi = new MenuItem(asName);
	mi.subMenu = splitMenu(items,null);
	mainMenu.addItem(mi);
	mi.subMenu.foreGround = mainMenu.foreGround;
	fixMenu(mi.subMenu);
	return mi.subMenu;
}
/**
 * Set a particular softkey to be a Menu, specifying the menu as a String.
 * @param which The key to set, starting from 1.
 * @param name The name of the menu.
 * @param icon An optional icon for the menu.
 * @param items The items for the menu. This should be specified either as a
 * comma separted list of item names (e.g. "Item 1,Item 2,Item 3") or as a '|' separated list of
 * item name and action pairs (e.g. "Item 1|ACTION_1|Item 2|ACTION_2|Item 3|ACTION_3").
 */
//===================================================================
public SoftKeyBar setMenu(int which, String name, IImage icon, String items)
//===================================================================
{
	Menu m = splitMenu(items,null);
	return setKey(which,name,icon,m);
}

//-------------------------------------------------------------------
protected Control make()
//-------------------------------------------------------------------
{
	Panel cp = new Panel(){
//		===================================================================
		public void layout(Grid controls,Panel panel,Rect panelRect)
//		===================================================================
		{
			if (isMobile5){
				int thick = preferredBarHeight >= 40 ? 4 : 2;
				int sw = grid.columns == 3 ? 
						//(SIPButton.sipActuallyOnRight ? SIPButton.sipButtonSize.width : 60) 
						SIPButton.sipButtonSize.width
						: 0;//SIPButton.sipButtonSize.width;
				if (isMobile6) sw = sw+2;
				//Device.messageBox("Size","Size: "+sw+", "+panelRect.width+", "+panelRect.height,Device.MB_TYPE_OK);
				//sw = panelRect.width/2;
				int first = (panelRect.width-sw)/2;
				int last = panelRect.width-first-sw;
				int h = panelRect.height-thick*2;
				int y = (panelRect.height-h)/2;
				for (int i = 0; i<grid.columns; i++){
					Control c = (Control)controls.objectAt(0,i);
					c.setBorder(c.BDR_NOBORDER,0);
					if (i == 0)
						c.setRect(0,y,first,h);
					else if (!SIPButton.sipActuallyOnRight) {
						if (i == grid.columns-1) c.setRect(first+sw,y,last,h);
						else {
							c.setRect(first,y,sw,h);
							if (isMobile6 && !SIPButton.simulate()){
								c.modify(0,c.Invisible);
								c.backGround = topStripe;
							}
						}
					}else{
						if (i == grid.columns-1) c.setRect(first,y,last,h);
						else c.setRect(first+last,y,sw,h);
					}
				}
			}else
				super.layout(controls,panel,panelRect);
		}
		public void doPaint(Graphics g, Rect where)
		{
			super.doPaint(g,where);
			int thick = preferredBarHeight >= 40 ? 4 : 2;
			int halfT = thick/2;
			if (isMobile6 && halfT != 0){
				g.setColor(Color.Black);
				g.fillRect(0,0,this.width,halfT);
				g.setColor(topStripe);
				g.fillRect(0,halfT,this.width,halfT);
			}else{
				g.setColor(topStripe);
				g.fillRect(0,0,this.width,thick);
			}
			if (isMobile6){
				//No bottom stripe.
			}else{
				g.setColor(bottomStripe);
				g.fillRect(0,this.height-thick,this.width,thick);
			}
		}
	};
	cp.setPreferredSize(-1,preferredBarHeight);
	cp.setFont(getFontMetrics().getFont());
	//if (keys.size() == 0) return cp;
	cp.equalWidths = true;
	int num = numberOfKeys();
	if (num < 2) num = 2;
	Color bc = Color.Black;
	boolean hasSIP = false;
	for (int i = 0; i<num; i++){
		Control ta = i > keys.size()-1 ? null : (Control)keys.get(i);
		if (ta == null) ta = new Label(" ");
		ta.borderColor = bc;
		cp.addNext(ta);
		if (i == 0) ta.setBorder(ta.BDR_OUTLINE|ta.BF_RIGHT|ta.BF_TOP|ta.BF_EXACT,1);
		else ta.setBorder(ta.BDR_OUTLINE|ta.BF_TOP|ta.BF_EXACT,0);
		if (i == 0) {
			hasSIP = SIPButton.placeInSoftKeyBar(cp);
		}
	}
	cp.endRow();
	cp.modifyAll(cp.NoFocus|barModifiersToSet,barModifiersToClear);
	cp.backGround = defaultBackgroundColor;
	cp.setBorder(cp.BDR_NOBORDER,0);
	cp.borderColor = bc;
	return cp;
}
//-------------------------------------------------------------------
private boolean handleKey(int which, KeyEvent ev)
//-------------------------------------------------------------------
{
	if (keys.size() >= which){
		Control c = (Control)keys.get(which-1);
		if ((c.modifiers & c.Disabled) != 0) return true;
		c.doAction(c.ByKeyboard);
		if (!c.menuIsActive()) c.notifyAction();
		return true;
	}
	return false;
}
//===================================================================
public boolean handleKey(KeyEvent ev)
//===================================================================
{
	if (ev.type == ev.KEY_RELEASE) return true;
	for (int i = 0; i<keys.size(); i++){
		Control c = (Control)keys.get(i);
		if (c.menuIsActive()) {
			c.closeMenu();
			return true;
		}
	}
	if (ev.key == IKeys.SOFTKEY1) return handleKey(1,ev);
	else if (ev.key == IKeys.SOFTKEY2) return handleKey(2,ev);
	else return false;
}
//===================================================================
public static boolean onKeyEvent(KeyEvent ev)
//===================================================================
{
	SoftKeyBar sk = getDisplayed();
	if (sk == null) return false;
	return sk.handleKey(ev);
}
//===================================================================
public void addTo(SingleContainer sc)
//===================================================================
{
	Control c = make();
	boolean changed = false;
	if (c == null) {
		changed = !sc.hasModifier(sc.ShrinkToNothing,false);
		sc.modify(sc.ShrinkToNothing,0);
		sc.setControl(new Label(" "));
	}else{
		changed = sc.hasModifier(sc.ShrinkToNothing,false);
		sc.modify(0,sc.ShrinkToNothing);
		sc.setControl(c);
	}
	if (changed){
		Control p = sc.getParent();
		if (p instanceof Panel){
			((Panel)p).relayoutMe(true);
		}
	}else
		sc.repaintNow();
	//if (p != null) p.redisplay();
}
/**
This returns true if SoftKeys are under this platform.
**/
//===================================================================
public static boolean usingSoftKeys()
//===================================================================
{
	return barContainer != null;
}
private static SingleContainer barContainer;
private static Frame nonSB, keypadFrame;
/**
* Setup the global location for the SoftKeyBar.
**/
//===================================================================
public static void setupIn(Panel c)
//===================================================================
{
	Menu.defaultBackground = Color.LightBlue;
	Menu.use3DPopup = false;
	c.addLast(barContainer = new SingleContainer());//.setCell(c.HSTRETCH);
	empty.addTo(barContainer);
}
static boolean hasKeypad()
{
	return keypadFrame != null;
}
static synchronized int keypadGetState()
{
	if (keypadVisible) return 1;
	return 0;
}
static boolean keypadVisible = false;
static synchronized void keypadSetVisible(boolean setVisible)
{
	if (!setVisible){
		if (!keypadVisible) return;
		keypadFrame.modify(keypadFrame.ShrinkToNothing,0);
		//new Exception().printStackTrace();
	}else{
		if (keypadVisible) return;
		keypadFrame.modify(0,keypadFrame.ShrinkToNothing);
	}
	keypadVisible = setVisible;
	//nonSB.relayout(true);
	nonSB.reShow(nonSB.x,nonSB.y,nonSB.width,nonSB.height);
	nonSB.getWindow().repaintNow();
	//try{Thread.sleep(2000);}catch(Exception e){}
	//nonSB.repaintNow();
}
static Frame getSofkeyControlMenuFrame(Control c)
{
	if (barContainer == null || nonSB == null) return null;
	if (c.isChildOf(barContainer)) return nonSB;
	return null;
}
private static SoftKeyBar current;
/**
Display the SoftKeyBar in the global SoftKeyBar location.
**/
//===================================================================
public boolean display()
//===================================================================
{
	reveal();
	if (current == this) return true;
	if (barContainer == null) return false;
 	addTo(barContainer);
	current = this;
	return true;
}
//===================================================================
public static SoftKeyBar getDisplayed()
//===================================================================
{
	if (barContainer == null) return null;
	if (barContainer.hasModifier(Control.ShrinkToNothing,false)) return null;
	return current;
}
static final int SETUP_SHOW_SOFTKEYS = 0x1;  
static final int SETUP_USE_KEYPAD = 0x2;
private static WindowSurface myWindow = null;
private static RawEvent re = new RawEvent();

private static void sendKeyEvent(int key,int mods)
{
	if (myWindow == null) myWindow = keypadFrame.getWindow().getWindowSurface(); 
	re.rawEventType = re.KEY_EVENT;
	re.key = key;
	re.modifiers = mods|IEventModifiers.POST_TO_BACK;
	re.surfaceEventType = SurfaceKeyEvent.KEY_PRESS;
	myWindow.postRawEvent(re);
}

/**
This is called by the GUI system when running on a Smartphone. It turns off
multiple windows and sets up a special area for the SoftKeyBar.
**/
//===================================================================
public static void setupScreen(int options)
//===================================================================
{
	if (isSingleMenuOnly()) showMenuNumbers = false;
	int h = Vm.getParameter(Vm.GET_PARAMATER_SOFTKEY_BAR_HEIGHT);
	if (h > 0) preferredBarHeight = h;
	defaultFont = Application.findFont("gui",true);
	if (preferredBarHeight >= 40){
		FontMetrics fm = Fx.getFontMetrics(defaultFont);
		defaultFont = Metrics.getFontForHeight((preferredBarHeight*5)/10,fm);
	}
	defaultFont = defaultFont.changeStyle(Font.BOLD);
	if (isMobile5){
		h = preferredBarHeight;
		
		if (h <= 26)
			SIPButton.sipButtonSize.width = 60;
		else if (h <= 35)
			SIPButton.sipButtonSize.width = 80;
		else if (h <= 52)
			SIPButton.sipButtonSize.width = 120;
		else
			SIPButton.sipButtonSize.width = (h*60)/26;
		
		//int wasWidth  = SIPButton.sipButtonSize.width;
		//SIPButton.sipButtonSize.width = (Gui.getScreenSize().width*263)/1000;
		//Device.messageBox("Size","Width is: "+SIPButton.sipButtonSize.width+", "+wasWidth,Device.MB_TYPE_OK);
	}
	/*
	final Form f = new Form();
	f.hasTopBar = false;
	f.windowFlagsToSet |= Window.FLAG_MAXIMIZE;
	Frame showFrame = new AppFrame();
	showFrame.setPreferredSize(100,100);
	f.addLast(showFrame);
	Frame ac = new Frame(){
		public void relayoutMe(boolean reDisp){
			super.relayoutMe(reDisp);
			f.relayoutMe(reDisp);
		}
	};
	ac.contentsOnly = true;
	ac.isControlPanel = true;
	ac.modify(ac.NoFocus,0);
	f.addLast(ac).setCell(f.HSTRETCH);
	SoftKeyBar.setupIn(ac);
	f.show();
	Gui.setAllParentFrame(showFrame);
	f.waitUntilPainted(1000);
	*/
	if (Gui.getNoMultipleWindows()){
		final Form f = new Form(){
			{
				windowTitle = WINDOW_TITLE_DONT_CHANGE;
				title = "All";
				dontAddSIPFiller = true;
			}
			protected void setupFrame(Frame ff,int options){
				ff.contentsOnly = true;
				//f.hasTopBar = false;
				ff.borderWidth = 0;
			}
			public boolean canExit(int value)
			{
				return false;
			}
			public void onEvent(Event ev){
				//super.onEvent(ev);
			}
		};
		f.backGround = Color.White;
		f.hasTopBar = false;
		f.exitButtonDefined = true;
		f.windowFlagsToSet |= Window.FLAG_MAXIMIZE;
		//
		// nonSB is everything out of the SoftkeyBar.
		//
		nonSB = new Frame();
		nonSB.contentsOnly = true;
		nonSB.isControlPanel = false;
		f.addLast(nonSB);
		//
		// Put the main app frame in the nonSB.
		//
		Frame showFrame = new AppFrame();
		showFrame.setPreferredSize(100,100);
		//f.addLast(showFrame);
		nonSB.addLast(showFrame);
		//
		// Now add a frame for the keypad.
		//
		if ((options & SETUP_USE_KEYPAD) != 0) 
		{
			Frame ac = new Frame(){
				public void relayoutMe(boolean reDisp){
					super.relayoutMe(reDisp);
					f.relayoutMe(reDisp);
				}
			};
			//f.addLast(new SIPKeypad());
			ac.contentsOnly = true;
			ac.isControlPanel = true;
			ac.modify(ac.NoFocus,0);
			nonSB.addLast(ac).setCell(f.HSTRETCH);
			ac.addLast(new SIPKeypad(0)).setCell(f.HSTRETCH).setControl(f.DONTFILL);
			keypadFrame = ac;
			keypadFrame.modify(keypadFrame.ShrinkToNothing,0);			
		}
		//
		//f.addLast(new SIPKeypad());
		if ((options & SETUP_SHOW_SOFTKEYS) != 0){
			Frame ac = new Frame(){
				public void relayoutMe(boolean reDisp){
					super.relayoutMe(reDisp);
					f.relayoutMe(reDisp);
				}
			};
			Gui.softkeyMadeFrame = ac;
			ac.contentsOnly = true;
			ac.isControlPanel = true;
			ac.modify(ac.NoFocus,0);
			if (isSingleMenuOnly() && 
					((Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_HAS_SOFTKEY_MENU_BUTTON) != 0)) 
				ac.setPreferredSize(0,0);
			f.addLast(ac).setCell(f.HSTRETCH);
			SoftKeyBar.setupIn(ac);
			//ac.addLast(new Button("Hello"));
		}
		f.show();
		Gui.setAllParentFrame(showFrame);
		f.waitUntilShown(1000);
		showFrame.makeFrameTopMost();
	}
}
/**
Hide any visible SoftKeyBar.
**/
//===================================================================
public static void hide()
//===================================================================
{
	if (!hideRemovesSoftKeyBar) empty.display();
	else{
		//new Exception().printStackTrace();
		if (barContainer == null) return;
		barContainer.modify(Control.ShrinkToNothing,0);
		Control p = barContainer.getParent();
		Window w = nonSB.getWindow();
		if (SIPButton.hasSipButton()){
			w.clearWindowFlags(w.FLAG_SHOW_SIP_BUTTON);
			Gui.setSIP(w, SIPConstants.SIP_REMOVE_BUTTON);
		}
		if (p instanceof Panel)((Panel)p).relayoutMe(false);
		w.repaintNow();
		//if (p != null) p.redisplay();
	}
}

/**
Reveal a hidden SoftKeyBar.
**/
//===================================================================
public static boolean reveal()
//===================================================================
{
	if (barContainer == null) return false;
	if ((barContainer.modifiers & barContainer.ShrinkToNothing) == 0) return true;
	barContainer.modify(0,Control.ShrinkToNothing);
	Control p = barContainer.getParent();
	if (p instanceof Panel)((Panel)p).relayoutMe(false);
	Window w = nonSB.getWindow();
	if (SIPButton.hasSipButton())
		w.setWindowFlags(w.FLAG_SHOW_SIP_BUTTON);
	w.repaintNow();
	return true;
}

private static Vector keybars;
/**
Display a new keybar, but save the old keybar.
**/
//===================================================================
public static void push(SoftKeyBar skb)
//===================================================================
{
	keybars = mVector.push(keybars,getDisplayed());
	if (skb == null) hide();
	else skb.display();
}
/**
Display the keybar that was displayed before the last push() call.
**/
//===================================================================
public static void pop()
//===================================================================
{
	SoftKeyBar got = (SoftKeyBar)mVector.pop(keybars);
	if (got == null) hide();
	else got.display();
}

//private static WeakSet fixedMenus;
/**
Fix a Menu for use with the SoftKeyBar. This will ensure that MenuItem objects are used
instead of Strings within the Menu and that numbers are prefixed to the labels.
**/
//===================================================================
public static Menu fixMenu(Menu m)
//===================================================================
{
	if (m == null) return m;
	m.exitOnLeft = true;
	//if (fixedMenus == null) fixedMenus = new WeakSet();
	//if (fixedMenus.contains(m)) return m;
	if (Gui.isSmartPhone)
		for (int i = 0, k = 0; i<m.items.size(); i++){
			Object obj = m.items.get(i);
			MenuItem mi = obj instanceof MenuItem ? (MenuItem)obj : new MenuItem(obj.toString());
			if (mi.action == null) mi.action = mi.label;
			if ((mi.modifiers & mi.SoftKeyFixed) == 0){
				mi.modifiers |= mi.SoftKeyFixed;
				if (showMenuNumbers && k <= 11 && !mi.label.equals("-") && ((mi.modifiers & mi.Separator) == 0)){
					k++;
					char ky = (char)('0'+k);
					if (k == 10) ky = '0';
					else if (k == 11) ky = '*';
					else if (k == 12) ky = '#';
					mi.label = ky+" "+mi.label;
					mi.hotkey = KeyEvent.toKey(0,ky);
					if (mi.image instanceof IconAndText){
						((IconAndText)mi.image).set(((IconAndText)mi.image).icon,mi.label);
					}
				}
			}
			m.items.setElementAt(mi,i);
		}
	//fixedMenus.add(m);
	return m;
}
//##################################################################
}
//##################################################################


