/*
 * Created on Oct 26, 2005
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.util;

import java.io.IOException;
import java.io.InputStream;

import eve.io.ByteArrayRandomStream;
import eve.io.File;
import eve.io.RandomStream;
import eve.io.RewindableStream;
import eve.io.StreamUtils;
import eve.sys.Handle;
import eve.sys.Vm;

/**
 * This class is used to provide bytes which represent some kind of formatted data.
 * The actual data format depends on the context it is used for. For example an
 * eve.fx.Picture() can be constructed using a FormattedDataSource, in which case
 * it is expected that the data bytes will be in a recognizable image format
 * (BMP, JPEG, GIF, PNG).
 * <p>
 * After constructing a new and empty FormattedDataSource use one of the set()
 * methods to specify the source of the data bytes. This can be a File, an 
 * already open InputStream, a named resource or bytes in memory. After calling
 * set() you can then read the byte data from the FormattedDataSource in a
 * variety of ways. You can call getInputStream()/getRandomStream() to get
 * the data as a stream, or you can call getByteArray()/getAllBytes() to
 * get the data as bytes in memory (this may involved reading in the data
 * from the source if the source was an InputStream).
 */
//####################################################
public class FormattedDataSource {

/** This is an optional name that you can give to the FormattedDataSource. */
public String name;

private byte[] src;
private int offset;
private int length;
private InputStream input;
private RandomStream rs;
private ByteArray ba;
private ArraySection as;

/**
 * Create a new FormattedDataSource. You can then use one of the set() methods
 * to assign data either from an InputStream or a set of bytes.
 */
public FormattedDataSource() {}

/**
 * If the FormattedDataSource is not already reading from bytes then read them all into
 * memory - closing the original data source.
 * @return this FormattedDataSource
 * @throws IOException
 */
public FormattedDataSource readIntoMemory() throws IOException
{
	if (src != null) return this;
	InputStream in = getInputStream();
	ByteArray ba = StreamUtils.readAllBytes(null,in,null);
	close();
	set(ba);
	return this;
}
/**
 * Clear the input source - but without closing any InputStreams. 
 */
public void clear()
{
	src = null;
	input = null;
	rs = null;
}
/**
 * This calls clear() but will also erase the internal ByteArray or ArraySection buffer.
 */
public void free()
{
	clear();
	ba = null;
	as = null;
}

/**
 * Close all open InputStreams and clear() the source, ignoring any
 * IOExceptions that may occur.
 */
public void close()
{
	try{if (input != null) input.close();}catch(IOException e){} 
	try{if (rs != null && rs != input) rs.close();}catch(IOException e){}
	clear();
}
/**
 * This sets the source of the bytes to be from the specified file.
 * @param src the file to retrieve the bytes from..
 * @return itself.
 * @throws IOException if the file could not be opened to read its bytes.
 */
public FormattedDataSource set(File src) throws IOException
{
	name = src.toString();
	return set(src.toReadableStream());
}
/**
 * This sets the source of the bytes to be from a named resource file or named resource data.
 * The method Vm.openResource(null,resourceName) is used to open the resource.
 * @param resourceName the name of the resource.
 * @return itself.
 * @throws IOException if the resource could not be found or opened.
 */
public FormattedDataSource set(String resourceName) throws IOException
{
	name = resourceName;
	return set(Vm.openResource(null,resourceName));
}
/**
 * This sets the source of the bytes to be an InputStream.
 * @param source the InputStream to read the formatted data from.
 * @return itself.
 */
public FormattedDataSource set(InputStream source)
{
	clear();
	input = source;
	return this;
}
/**
 * This sets the source of the bytes to be from an array of bytes.
 * @param data the array containing the bytes.
 * @param offset the start offset within the array.
 * @param length the number of bytes in the data.
 * @return itself.
 */
public FormattedDataSource set(byte[] data, int offset, int length)
{
	clear();
	src = data;
	this.offset = offset;
	this.length = length;
	return this;
}
/**
 * This sets the source of the bytes to be from an array of bytes.
 * @param data the ByteArray containing the bytes.
 * @return itself.
 */
public FormattedDataSource set(ByteArray data)
{
	clear();
	src = data.data;
	this.offset = 0;
	this.length = data.length;
	return this;
}
/**
 * This checks to see if one of the set() methods has been called.
 * @return true if there is data to be read, false if not.
 */
public boolean hasData()
{
	return input != null && src != null;
}

/**
 * Return an InputStream for reading the formatted data bytes. 
 * If the set(InputStream) method was used then this returns that InputStream.
 * If the set(byte[],int,int) method was used then a ByteArrayRandomStream is returned.
 * @return an InputStream for reading the formatted data bytes.
 */
public synchronized InputStream getInputStream()
{
	if (input == null) return input = new ByteArrayRandomStream(src,offset,length,"r");
	if (rs != null) return rs;
	return input;
}
/**
 * Return a RewindableStream for reading the data bytes. Calling rewind() afterwards
 * will reset the stream so that all further reads will read from the beginning of
 * the stream. Note however that RewindableStreams can be rewinded only once.
 * @return a RewindableStream for reading the data bytes.
 */
public synchronized RandomStream getRewindableStream()
{
	if (rs == null) rs = RewindableStream.toRewindableStream(getInputStream());
	return rs;
}
/**
 * Get a true RandomStream to read the formatted data bytes from. This may involve
 * reading all the bytes into memory first.
 * @return a RandomStream to read the formatted data bytes.
 * @throws IOException if there is an IO error reading the data bytes into memory 
 * if that was necessary.
 */
public synchronized RandomStream getRandomStream() throws IOException
{
	InputStream is = getInputStream();
	if (rs == null && is instanceof RandomStream && !(is instanceof RewindableStream))
		rs = (RandomStream)is;
	if (rs != null && !(rs instanceof RewindableStream)) return rs;
	if (src == null) {
		ByteArray ba = StreamUtils.readAllBytes(null,is,null);
		src = ba.data;
		offset = 0;
		length = ba.length;
	}
	if (src != null) {
		if (input != null) input.close();
		input = rs = new ByteArrayRandomStream(src,offset,length,"r");		
	}
	return rs;
}
/**
 * Call this only once after calling getRewindableStream(). It will cause the RandomStream
 * to return bytes from the beginning on the next read operation.
 * @throws IOException if an IO error occurs.
 * @throws IllegalStateException if getRewindableStream() was not called first.
 */
public synchronized void rewind() throws IOException, IllegalStateException
{
	if (rs == null) throw new IllegalStateException();
	RewindableStream.rewind(rs);
}

/**
 * Read in all the bytes and place in a ByteArray. If the set(byte[],int,int) method
 * was used to set the source then this copies the bytes directly into the ByteArray.
 * @param handle an optional Handle that can be used to stop the process if another
 * Thread calls stop() on the Handle.
 * @param destination an optional destination ByteArray. If this is null a new one
 * will be created and returned.
 * @return the destination ByteArray or a new one if it was null. If stop() was called
 * on the Handle then null will be returned.
 * @throws IOException if an IO error occured reading in the bytes.
 */
public ByteArray getByteArray(Handle handle,ByteArray destination) throws IOException
{
	if (destination == null) destination = new ByteArray();
	if (src == null)
		return StreamUtils.readAllBytes(handle,getInputStream(),destination);
	destination.append(src,offset,length);
	return destination;
}
/**
 * Read in all the bytes and place in a ByteArray. If the set(byte[],int,int) method
 * was used to set the source then this copies the bytes directly into the ByteArray.
 * @param handle an optional Handle that can be used to stop the process if another
 * Thread calls stop() on the Handle.
 * @return a ByteArray containing the data or null if stop() was called on the Handle.
 * The ByteArray returned will be re-used if getByteArray() is called again.
 * @throws IOException if an IO error occurs reading in the bytes.
 */
public ByteArray getByteArray(Handle handle) throws IOException
{
	if (ba == null) ba = new ByteArray();
	else ba.clear();
	return getByteArray(handle,ba);
}

/**
 * Read in all the data, place it in a byte array and then place the result in an
 * ArraySection.  
 * If the set(byte[],int,int) method
 * was used to set the source then that data is placed in the ArraySection without
 * copying of data.
 * @param handle an optional Handle that can be used to stop the process if another
 * Thread calls stop() on the Handle.
 * @param destination an optional destination ArraySection. If this is null a new one
 * will be created and returned.
 * @return the destination ArraySection or a new one if it was null. If stop() was called
 * on the Handle then null will be returned.
 * @throws IOException if an IO error occurs reading in the bytes.
 */
public ArraySection getAllBytes(Handle handle, ArraySection destination) throws IOException
{
	if (destination == null) destination = new ArraySection();
	if (src == null) {
		ByteArray ba = getByteArray(handle);
		if (ba == null) return null;
		destination.data = ba.data;
		destination.startIndex = 0;
		destination.length = ba.length;
		return destination;
	}
	destination.data = src;
	destination.length = length;
	destination.startIndex = offset;
	return destination;
}
/**
 * Read in all the data, place it in a byte array and then place the result in an
 * ArraySection.  
 * If the set(byte[],int,int) method
 * was used to set the source then that data is placed in the ArraySection without
 * copying of data.
 * @param handle an optional Handle that can be used to stop the process if another
 * Thread calls stop() on the Handle.
 * @return an ArraySection containing the data or null if stop() was called on the Handle.
 * The ArraySection returned will be re-used if getAllBytes() is called again.
 * @throws IOException if an IO error occurs reading in the bytes.
 */
public ArraySection getAllBytes(Handle handle) throws IOException
{
	if (as == null) as = new ArraySection();
	else as.clear();
	return getAllBytes(handle,as);
}
/**
 * Expose the source bytes as an ArraySection ONLY if the data was provided as an
 * array of bytes. If the data is being provided via an InputStream then this will
 * return null.
 * @return an ArraySection holding the source bytes only if the data was provided
 * as an array of bytes. Returns null otherwise.
 */
public ArraySection exposeBytes()
{
	if (src == null) return null;
	if (as == null) as = new ArraySection();
	as.clear();
	as.data = src;
	as.length = length;
	as.startIndex = offset;
	return as;
}
}
//####################################################
