/*
 * Created on Oct 7, 2006
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.util;

import java.util.Vector;

import eve.sys.TimeOut;

/**
 * This is used to pass Object data from one Thread to another safely
 * without leaving unhandled data.
 */
//####################################################
public class ObjectBuffer 
{
	private boolean closed = false;
	private Vector objects;// = new Vector();
	/**
	 * Create an ObjectBuffer with a maximum size. If this size is negative
	 * then there is no maximum size.
	 * @param maxSize the maximum number of Objects the buffer can hold.
	 */
	public ObjectBuffer(int maxSize)
	{
		this.maxSize = maxSize;
	}
	/**
	 * Create an ObjectBuffer with no maximum size.
	 */
	public ObjectBuffer()
	{
		
	}
	public int maxSize = -1;
	/**
	 * This is used to discard an Object that was in the buffer but not taken
	 * by the time the buffer closed.
	 * @param obj the object.
	 */
	/*
	protected void discard(Object obj)
	{
		
	}
	*/
	
	/**
	 * Returns if the buffer is closed so no more data can be put into it. If it is
	 * closed but data remains in it that data can still be taken out. 
	 */
	public synchronized boolean isClosed()
	{
		return closed;
	}
	/**
	 * This closes the buffer so no more data can be put in it - however
	 * data can still be taken out of it.
	 */
	public synchronized void close()
	{
		if (closed) return;
		closed = true;
		notifyAll();
		closed();
	}
	/**
	 * This is called when the buffer is closed. It is synchronized with this Object.
	 */
	protected void closed()
	{
	}
	/**
	 * Get an Object from the buffer. 
	 * @param howLong the length of time to wait for the next object.
	 * @return the Object or null if there was no Object ready or if the TimeOut
	 * expired. Use isClosed() to determine if the buffer is closed. 
	 * @throws InterruptedException if the Thread was interrupted.
	 */
	public synchronized Object get(TimeOut howLong) throws InterruptedException
	{
		return get(howLong,false);
	}
	/**
	 * Get an Object from the buffer. 
	 * @param howLong the length of time to wait for the next object.
	 * @param closeIfFailed set this true if you want to close the buffer
	 * if data does not arrive in time or if ther Thread is interrupted.
	 * @return the Object or null if there was no Object ready or if the TimeOut
	 * expired. Use isClosed() to determine if the buffer is closed. 
	 * @throws InterruptedException if the Thread was interrupted.
	 */
	public synchronized Object get(TimeOut howLong, boolean closeIfFailed) throws InterruptedException
	{
		if (howLong == null) howLong = TimeOut.Forever;
		while(true){
			if (objects != null && objects.size() != 0){
				Object ret = objects.get(0);
				objects.removeElementAt(0);
				notifyAll();
				return ret;
			}
			if (howLong.hasExpired() || closed) {
				if (closeIfFailed) close();
				return null;
			}
			try{
				howLong.waitOn(this);
			}catch(InterruptedException e){
				if (closeIfFailed) close();
				throw e;
			}
		}
	}
	/**
	 * Place data in the buffer waiting if the buffer is full.
	 * @param data the data to put in the buffer.
	 * @param howLong how long to wait.
	 * @return true if the data was placed in the buffer, false if not because
	 * the buffer is full or the buffer is closed. Use isClosed() to determine if
	 * the buffer is closed.
	 */
	public synchronized boolean put(Object data, TimeOut howLong) throws InterruptedException
	{
		if (howLong == null) howLong = TimeOut.Forever;
		while(true){
			if (closed) break;
			if (objects == null) objects = new Vector();
			if (maxSize >= 0 && objects.size() < maxSize){
				objects.add(data);
				notifyAll();
				return true;
			}
			if (howLong.hasExpired()) break;
			howLong.waitOn(this);
		}
		return false;
	}
	/**
	 * Place data in the buffer but do not wait if the buffer is full.
	 * @param data the data to put in the buffer.
	 * @return true if the data was placed in the buffer, false if not because
	 * the buffer is full or the buffer is closed. Use isClosed() to determine if
	 * the buffer is closed.
	 */
	public synchronized boolean put(Object data)
	{
		try{
			return put(data,TimeOut.Immediate);
		}catch(InterruptedException e){
			return false;
		}
	}
}

//####################################################
