package eve.util;
import java.util.Vector;

import eve.data.DataObject;
import eve.sys.Cache;
import eve.sys.Locale;
import eve.sys.Vm;
/**

This is a very specialized Utility class used for bulk processing of
text strings, while minimizing overheads of new Object creation.<p>
A SubString consists of a reference to a character array 
(which may be shared by other SubStrings), a start index within the array and a 
length value for the sub-string.<p>
You can manipulate a SubString in many of the same ways as you do a String, but
unlike a String, a SubString is mutable (since you have full access to its data
and its start and length values).

**/
//##################################################################
public class SubString 
extends DataObject{
//implements Comparable{
//##################################################################
public char [] data = null;
public int start = 0;
public int length = 0;
/*.....................................................*/
/**
 * This returns a new SubString().
 */
public Object getNew()
{
	return new SubString();
}
/**
* Return a String representation of this SubString.
**/
//===================================================================
public String toString()
//===================================================================
{
	if (data == null) return null;
	return new String(data,start,length);
}
/**
 * Return if this SubString is null (i.e. it's data is null).
 */
//===================================================================
public boolean isNull() {return data == null;}
//===================================================================
/**
* Get the length of the SubString.
**/
//===================================================================
public int length() {return length;}
//===================================================================
/**
* Return the character at a particular index.
**/
//===================================================================
public char charAt(int idx) {return data[idx+start];}
//===================================================================
/**
* Copy all the characters of this substring into a character array.
* @param dest The destination character array.
* @param offset The offset in the destination to start copying into.
*/
//===================================================================
public void copyInto(char[] dest,int offset)
//===================================================================
{
	copyInto(dest,offset,length,0);
}
/**
 * Copy a subset of this SubStrings data into a character array.
* @param dest The destination character array.
* @param offset The offset in the destination to start copying into.
 * @param len The number of characters to copy.
 * @param startPoint The first index in this substring to copy from.
 */
//===================================================================
public void copyInto(char[] dest,int offset,int len,int startPoint)
//===================================================================
{
	if (data == null) return;
	System.arraycopy(data,startPoint+start,dest,offset,len);
}
/**
* Copy all the characters of this substring into a byte array.
* @param dest The destination byte array.
* @param offset The offset in the destination to start copying into.
* @param ascii if this is true then only the low byte is copied into the destination.
* If you want to do Utf8 encoding, then see the encode method.
*/
//===================================================================
public void copyInto(byte[] dest,int offset,boolean ascii)
//===================================================================
{
	copyInto(dest,offset,length,0,ascii);
}
/**
 * Copy a subset of this SubStrings data into a byte array.
* @param dest The destination byte array.
* @param offset The offset in the destination to start copying into.
 * @param len The number of characters to copy.
 * @param startPoint The first index in this substring to copy from.
* @param ascii if this is true then only the low byte is copied into the destination.
* If you want to do Utf8 encoding, then see the encode method.
*/
//===================================================================
public void copyInto(byte [] dest,int offset,int len,int startPoint,boolean ascii)
//===================================================================
{
	if (data == null) return;
	if (ascii)
		for (int i = 0; i<len; i++)
			dest[i+offset] = (byte)(data[start+startPoint+i]&0xff);
	else
		for (int i = 0; i<len; i++){
			dest[i*2+offset] = (byte)((data[start+startPoint+i] >> 8)&0xff);
			dest[i*2+offset+1] = (byte)(data[start+startPoint+i]&0xff);
	}	
}
//===================================================================
public SubString set(char [] dt,int st,int len)
//===================================================================
{
	data = dt; start = st; length = len;
	return this;
}
//===================================================================
public SubString set(String string)
//===================================================================
{
	if (string == null) return set(string,0,0);
	else return set(string,0,string.length());
}
/**
* Create a SubString based on a substring of the text in the provided String. 
* A copy of the String data is made for this SubString.
**/
//===================================================================
public SubString set(String string,int st,int len)
//===================================================================
{
	if (string != null){
		data = new char[len];
		start = 0;
		length = len;
		string.getChars(st,st+len,data,0);
	}else{
		data = null;
		start = length = 0;
	}
	return this;
}
/**
* Create a new SubString - use one of the set() methods to setup the
* text for the SubString.
**/
//===================================================================
public SubString(){}
//===================================================================
/**
* This will split this substring and place the results as a series of
* SubString objects in the destination list. If the destination contains
* SubString objects already, then each will be used in sequence as a destination. If there
* are not enough SubStrings in the destination then more will be added. If
* there are too many SubStrings extra ones will NOT be removed. The return value is the number
* of substrings found.
* @param separator The separator character. A separator of ' ' (a space) will
* also use the TAB ('\t') as a separator. A space separator also has the property
* that a sequence of more than one space or TAB characters will be considered as just one
* separator. So a line like "hello    there how    are you" will be separated into the five
* words exactly.
* @param destination This is the destination vector (which must NOT be null) in which SubString
* objects representing each separated text is placed.
* @return The number of found SubStrings in the text.
*/
//===================================================================
public int split(char separator,java.util.Vector destination)
//===================================================================
{
	return split(separator,destination,0);
}
/**
* This will split this substring and place the results as a series of
* SubString objects in the destination list. If the destination contains
* SubString objects already, then each will be used in sequence as a destination. If there
* are not enough SubStrings in the destination then more will be added, taking from
* the used Vector if available. If
* there are too many SubStrings, extra ones will NOT be removed. The return value is the number
* of substrings found.
* @param separator The separator character. A separator of ' ' (a space) will
* also use the TAB ('\t') as a separator. A space separator also has the property
* that a sequence of more than one space or TAB characters will be considered as just one
* separator. So a line like "hello    there how    are you" will be separated into the five
* words exactly.
* @param destination This is the destination vector (which must NOT be null) in which SubString
* objects representing each separated text is placed.
* @param destinationStart. This tells which index in the destination Vector to start placing
* SubString values into.
* @param used an optional array containing SubStrings that may be re-used.
* @return The number of found SubStrings in the text.
*/
//===================================================================
public int split(char separator,Vector destination,int destinationStart,Vector used)
//===================================================================
{
	boolean space = separator == ' ';
	if (data == null) return 0;
	if (length == 0) return 0;
	int found = 0;
	int end = start+length;
	int usz = used == null ? 0 : used.size();
	if (space) {
		for (int i = start; i<end;){
			while(i<end && (data[i] == ' ' || data[i] == '\t')) i++;
			if (i>=end) break;
			int st = i;
			while(i<end && (data[i] != ' ' && data[i] != '\t')) i++;
			SubString d;
			if (found+destinationStart >= destination.size()) {
				if (usz != 0){
					--usz;
					d = (SubString)used.get(usz);
					used.removeElementAt(usz);
				}else
					d = (SubString)getNew();
				destination.add(d);
			}else 
				d = (SubString)destination.get(found+destinationStart);
			d.set(data,st,i-st);
			found++;
		}
	}else{
		for (int i = start;;){
			int st = i;
			while(i<end && data[i] != separator) i++;
			SubString d;
			if (found+destinationStart >= destination.size()) {
				if (usz != 0){
					--usz;
					d = (SubString)used.get(usz);
					used.removeElementAt(usz);
				}else
					d = (SubString)getNew();
				destination.add(d);
			}else d = (SubString)destination.get(found+destinationStart);
			d.set(data,st,i-st);
			found++;
			if (i>=end) break;
			i++;
		}
	}
	return found;
}
/**
* This will split this substring and place the results as a series of
* SubString objects in the destination list. If the destination contains
* SubString objects already, then each will be used in sequence as a destination. If there
* are not enough SubStrings in the destination then more will be added. If
* there are too many SubStrings, extra ones will NOT be removed. The return value is the number
* of substrings found.
* @param separator The separator character. A separator of ' ' (a space) will
* also use the TAB ('\t') as a separator. A space separator also has the property
* that a sequence of more than one space or TAB characters will be considered as just one
* separator. So a line like "hello    there how    are you" will be separated into the five
* words exactly.
* @param destination This is the destination vector (which must NOT be null) in which SubString
* objects representing each separated text is placed.
* @param destinationStart. This tells which index in the destination Vector to start placing
* SubString values into.
* @return The number of found SubStrings in the text.
*/
//===================================================================
public int split(char separator,Vector destination,int destinationStart)
//===================================================================
{
	return split(separator,destination,destinationStart,null);
}

//private static Vector strs = new Vector();


/**
 * Split a String into SubStrings placed in an output Vector.
 * @param what The String to split.
 * @param separator The separator character.
 * @param dest A destination Vector (which may be null).
 * @return A Vector containing the exact number of Strings found in the String.
 */
//===================================================================
public static java.util.Vector split(String what,char separator,java.util.Vector dest)
//===================================================================
{
	if (dest == null) dest = new mVector();
	dest.removeAllElements();
	SubString ss = new SubString().set(what);
	int got = ss.split(separator,dest,0);
	for (int i = 0; i<got; i++)
		dest.setElementAt(dest.elementAt(i).toString(),i);
	return dest;
}

/**
* An option for equals().
**/
public static final int IGNORE_CASE = 0x1;
/**
* An option for equals().
**/
public static final int STARTS_WITH = 0x2;
/**
* An option for equals().
**/
public static final int BACKWARDS = 0x4; //Sometimes a backwards check is faster.
/**
* This compares the character sequence given by the "small" parameters 
* to the character sequence given by the "big" parameters. 
**/
//===================================================================
public static boolean equals(
	char [] big,int bigStart,int bigLength,
	char [] small,int smallStart,int smallLength,
	int options)
//===================================================================
{
	boolean sw = (options & STARTS_WITH) != 0;
	boolean ig = (options & IGNORE_CASE) != 0;
	boolean back = (options & BACKWARDS) != 0;
	
	if (bigLength < smallLength) return false;
	if (!sw && (bigLength != smallLength)) return false;
	if (!back){
		int s = smallStart, end = smallLength+smallStart;
		int b = bigStart;
		if (!ig){
			while(s<end)
				if (small[s++] != big[b++]) return false;
		}else{
			while(s<end){
				char c1 = small[s++];
				char c2 = big[b++];
				if (c1 >= 'a' && c1 <= 'z') c1 ^= 0x20;
				if (c2 >= 'a' && c2 <= 'z') c2 ^= 0x20;
				if (c1 != c2) return false;
			}
		}
	}else{
		int s = smallStart+smallLength-1, end = smallStart;
		int b = bigStart+smallLength-1;
		if (!ig){
			while(s>=end)
				if (small[s--] != big[b--]) return false;
		}else{
			while(s>=end){
				char c1 = small[s--];
				char c2 = big[b--];
				if (c1 >= 'a' && c1 <= 'z') c1 ^= 0x20;
				if (c2 >= 'a' && c2 <= 'z') c2 ^= 0x20;
				if (c1 != c2) return false;
			}
		}
	}
	return true;
}
/**
 * Find the index of a string within another string.
 * @param other the characters of the smaller String.
 * @param otherStart the start of the smaller String in the array.
 * @param otherLength the number of characters in the smaller String.
 * @param inStr the larger String to search within.
 * @param strStart the start of the larger String characters in the array.
 * @param strLength the number of characters in the larger String.
 * @param start the first location to start looking from (this must be absolute and
 * not relative to strStart).
 * @return the index of the smaller String in the larger String if present.
 */
public static int indexOf(char[] other,int otherStart, int otherLength, char[] inStr, int strStart, int strLength, int start)
//===================================================================
{
	if (other == null) throw new NullPointerException();
	if (otherLength <= 0) return start;
	if (otherLength > strLength) return -1;
	int max = strStart+strLength;
	for (int s = start; s<max; s++){
		int idx = SubString.indexOf(other[otherStart],inStr,strStart,strLength,s,0);
		if (idx == -1) return -1;
		if (SubString.equals(inStr,idx,max-idx,other,otherStart,otherLength,SubString.STARTS_WITH)) return idx;
		s = idx; // s++ will add one to this.
	}
	return -1;
}
/**
 * Find the index of a string within another string.
 * @param other the characters of the smaller String.
 * @param otherStart the start of the smaller String in the array.
 * @param otherLength the number of characters in the smaller String.
 * @param inStr the larger String to search within.
 * @return the index of the smaller String in the larger String if present.
 */
public static int indexOf(char[] subString,int ssStart,int ssLength,String inStr)
{
	return indexOf(subString,ssStart,ssLength,Vm.getStringChars(inStr),0,inStr.length(),0);
}

//===================================================================
public static int indexOf(char what,char [] str,int strStart,int strLength,int start,int options)
//===================================================================
{
	if ((options & BACKWARDS) == 0){
		int s = start, end = strStart+strLength;
		while(s<end)
			if (str[s++] == what) return s-1;
		return -1;
	}else{
		int s = start, end = strStart;
		while(s>=end)
			if (str[s--] == what) return s+1;
		return -1;
	}
}
/**
* This does a numeric char for char comparison.
**/
//===================================================================
public static int compare(char [] one,int oneStart,int oneLength,char [] two,int twoStart,int twoLength)
//===================================================================
{
	if (one == two) return 0;
	if (one == null) return -1;
	else if (two == null) return 1;
	
	int max = oneLength > twoLength ? twoLength : oneLength;
	for (int i = 0; i<max; i++)
		if (one[oneStart+i] > two[twoStart+i]) return 1;
		else if (one[oneStart+i] < two[twoStart+i]) return -1;
	if (oneLength > twoLength) return 1;
	else if (oneLength < twoLength) return -1;
	else return 0;
}
/**
* Compare this SubString with either another SubString or a String.
* @param other Either another SubString or a String.
* @return greater than 1 if this SubString is considered greater than the Object, 
* less than 1 if this SubString is considered less than the Object,  and 0 if this
* SubString is considered equal to the Object.
**/
//===================================================================
public int compareTo(Object other)
//===================================================================
{
	return compareTo(other,0);
}
/**
* Compare this SubString with either another SubString or a String.
* @param other Either another SubString or a String.
* @param options This can be IGNORE_CASE or 0.
* @return greater than 1 if this SubString is considered greater than the Object, 
* less than 1 if this SubString is considered less than the Object,  and 0 if this
* SubString is considered equal to the Object.
*/
//===================================================================
public int compareTo(Object other,int options)
//===================================================================
{
	int opt = options == IGNORE_CASE ? Locale.IGNORE_CASE : 0;
	if (other instanceof SubString){
		SubString o = (SubString)other;
		if (data == null)
			if (o.data == null) return 0;
			else return -1;
		return Vm.getLocale().compare(data,start,length,o.data,o.start,o.length,opt);
	}
	if (data == null) return -1;
	if (other instanceof String){
		char [] c = Vm.getStringChars((String)other);
		return Vm.getLocale().compare(data,start,length,c,0,c.length,opt);
	}else
		return 1;
}
/**
* As specified in the hashCode() contract, this returns the same hashCode that a String
* with the same text would return.
**/
//===================================================================
public int hashCode()
//===================================================================
{
	if (data == null) return 0;
	return Utils.makeHashCode(data,start,length);
}
/**
 * Get a new SubString that is a substring of this SubString, optionally copying the data
 * out into a new array.
 * @param start The first index in this SubString.
 * @param length The number of characters.
 * @param destination The destination SubString. If this is null a new one will be created.
 * @param copyDataInto If this is null then the new SubString will reference the original
 * character array for its data. If it is not null then the characters will be copied into
 * this array. if this array is not big enough, a new one will be created and that one will
 * be used instead.
 * @return A new SubString (or the provided destination) representing part or all of the original
 * substring.
 */
//===================================================================
public SubString substring(int start,int length,SubString destination,char[] copyDataInto)
//===================================================================
{
	if (start < 0 || start+length > this.length) throw new IndexOutOfBoundsException();
	if (destination == null) destination = (SubString)getNew();
	if (copyDataInto != null){
		if (copyDataInto.length < length) copyDataInto = new char[length];
		System.arraycopy(data,this.start+start,copyDataInto,0,length);
		destination.data = copyDataInto;
		destination.start = 0;
		destination.length = length;
	}else{
		destination.data = data;
		destination.start = this.start+start;
		destination.length = length;
	}
	return destination;
}
//===================================================================
public void toUpperCase()
//===================================================================
{
	if (data == null) return;
	Vm.getLocale().changeCase(data,start,length,true);
}
//===================================================================
public void toLowerCase()
//===================================================================
{
	if (data == null) return;
	Vm.getLocale().changeCase(data,start,length,false);
}
/**
 * Get a new SubString that is a substring of this SubString.
 * @param start The first index in this SubString.
 * @param length The number of characters.
 * @param destination The destination SubString. If this is null a new one will be created.
 * @return A new SubString (or the provided destination) representing part or all of the original
 * substring.
 */
//===================================================================
public SubString substring(int start,int length,SubString destination)
//===================================================================
{
	return substring(start,length,destination,null);
}
/**
 * Get a new SubString that is a substring of this SubString.
 * @param start The first index in this SubString.
 * @param length The number of characters.
 * @return A new String representing part or all of the original
 * substring.
 */
//===================================================================
public String substring(int start,int length)
//===================================================================
{
	if (data == null) return null;
	return new String(data,start+this.start,length);
}
/**
* Get a copy of the SubString.
 * @param destination The destination SubString. If this is null a new one will be created.
 * @param copyDataInto If this is null then the new SubString will reference the original
 * character array for its data. If it is not null then the characters will be copied into
 * this array. if this array is not big enough, a new one will be created and that one will
 * be used instead.
 * @return A new SubString (or the provided destination) representing part or all of the original
 * substring.
**/
//===================================================================
public SubString getCopy(SubString destination,char[] copyDataInto)
//===================================================================
{
	return substring(0,length,destination,copyDataInto);
}

/*
{
	if (one == two) return 0;
	if (one == null) return -1;
	else if (two == null) return 1;
	
	int max = oneLength > twoLength ? twoLength : oneLength;
	for (int i = 0; i<max; i++)
		if (one[oneStart+i] > two[twoStart+i]) return 1;
		else if (one[oneStart+i] < two[twoStart+i]) return -1;
	if (oneLength > twoLength) return 1;
	else if (oneLength < twoLength) return -1;
	else return 0;
}
*/

/**
* Check if the SubString startsWith a particular String.
**/
//===================================================================
public boolean startsWith(String startWith)
//===================================================================
{
	if (data == null) return false;
	char [] sc = Vm.getStringChars(startWith);
	int len = sc.length;
	if (len > length) return false;
	return Vm.getLocale().compare(data,start,len,sc,0,len,0) == 0;
}
/**
* Check if the SubString endsWith a particular String.
**/
//===================================================================
public boolean endsWith(String endWith)
//===================================================================
{
	if (data == null) return false;
	char [] ec = Vm.getStringChars(endWith);
	int len = ec.length;
	if (len > length) return false;
	return Vm.getLocale().compare(data,start+length-len,len,ec,0,len,0) == 0;
}
/**
* Trim the SubString so there are no leading or trailing spaces or Tabs.
**/
//===================================================================
public void trim()
//===================================================================
{
	if (data == null) return;
	for (;length > 0;length--)
		if (data[start] != ' ' && data[start] != '\t')
			break;
		else
			start++;
	for (;length > 0;length--)
		if (data[start+length-1] != ' ' && data[start+length-1] != '\t')
			break;
}
//===================================================================
public void replace(char[] what,char[] with)
//===================================================================
{
	if (what.length == 0) return;
	int whatLen = what.length;
	int withLen = with.length;
	int dif = withLen-whatLen;
	int max = start+length-(whatLen-1);
	for (int s = start; s<max; s++){
		if (data[s] == what[0]){
			boolean found = true;
			if (whatLen != 1)
				for(int i = 1; i<whatLen; i++){
					if (data[s+i] != what[i]){
						found = false;
						break;
					}
				}
			if (!found) continue;
			//
			// Expand char array if necessary.
			//
			if (dif > 0 && start+length+dif > data.length){
				char [] newChar = new char[start+length+dif*20];
				System.arraycopy(data,start,newChar,start,length);
				data = newChar;
			}
			//
			// Move data.
			//
			if (dif != 0)
				System.arraycopy(data,s+whatLen,data,s+withLen,length-(s+whatLen-start));
			//
			// Copy first character.
			//
			if (withLen == 1) data[s] = with[0];
			//
			// Copy others.
			//
			else if (withLen > 1) System.arraycopy(with,0,data,s,withLen);
			//
			length += dif;
			max += dif;
			//s--;
			// This bypasses the new characters added.
			s += withLen-1;
		}
	}
}
public SubString set(CharArray ca,int start,int end)
{
	return set(ca.data,start,end);
}
public SubString set(CharArray ca)
{
	return set(ca.data,0,ca.length);
}
public static SubString getCached()
{
	return (SubString)Cache.get(SubString.class);
}

/**
 * Convert a String to a SubString without making a copy of the String characters if
 * possible.
 * @param s the String to access the char data of.
 * @return this SubString
 */
public SubString getStringChars(String s)
{
	data = Vm.getStringChars(s);
	start = 0;
	length = s.length();
	return this;
}
public void cache()
{
	Cache.put(this);
}
public void cached()
{
	data = null;
}

//===================================================================
public int indexOf(int ch) {return indexOf(ch,0);}
public int indexOf(int ch,int start)
//===================================================================
{
	return SubString.indexOf((char)ch,data,this.start,length,start+this.start,0);
}
//===================================================================
public int lastIndexOf(int ch) {return lastIndexOf(ch,length-1);}
public int lastIndexOf(int ch,int start)
//===================================================================
{
	return SubString.indexOf((char)ch,data,this.start,length,this.start+start,SubString.BACKWARDS);
}

//##################################################################
}
//##################################################################


