package eve.util;
import eve.sys.Type;
import eve.sys.WeakSet;
import eve.sys.mSecurityManager;
import eve.data.PropertyList;
import eve.io.File;
import eve.io.StreamUtils;
import eve.io.filestore.FileStore;
import eve.io.filestore.FileStoreEntry;

import java.io.IOException;
import java.io.InputStream;
import java.util.Hashtable;
import java.util.Vector;
//##################################################################
public class mClassLoader extends ClassLoader{
//##################################################################

private static WeakSet classLoaders;

/**
 * If you set this to a directory File, then the default behavior of the
 * mClass loader is to try to load all classes from this File.
 */
public File sourceDirectory;
/**
 * If you set this to a non-null FileStore, then the default behavior of the
 * mClass loader is to try to load all classes from this FileStore.
 */
public FileStore fileStore;

/**
* This passes authority from this ClassLoader to another. This call
* can only be made by system classes or by classes loaded by this ClassLoader,
* or by this ClassLoader itself.
**/
//===================================================================
public final void authorize(mClassLoader other) throws SecurityException
//===================================================================
{
	SecurityManager sm = System.getSecurityManager();
	if (sm instanceof mSecurityManager){
		mSecurityManager ms = ((mSecurityManager)sm);
		ms.checkPassAuthority(this,other);
		PropertyList mine = null;
		try{
			mine = ms.getAuthorizers(this,false);
		}catch(SecurityException e){
		// Nothing to authorize.
			return;
		}
		PropertyList pl = ms.getAuthorizers(other,true);
		pl.addAll(mine);
		return;
	}
	throw new SecurityException();
}
/**
* Add a ClassLoader to the WeakSet of application classLoaders. You do not need
* to do this if you are creating a ClassLoader that inherits from mClassLoader
* as they automatically add to the WeakSet upon creation.
**/
//===================================================================
public static synchronized void addClassLoader(ClassLoader loader)
//===================================================================
{
	if (loader == null) return;
	if (classLoaders == null) classLoaders = new WeakSet();
	classLoaders.add(loader);
}
/**
* Remove a ClassLoader from the WeakSet of application classLoaders. 
**/
//===================================================================
public static synchronized void removeClassLoader(ClassLoader loader)
//===================================================================
{
	if (classLoaders == null) return;
	classLoaders.remove(loader);
	if (classLoaders.isEmpty()) classLoaders = null;
}
/**
 * Get the current set of application ClassLoaders into a Vector, or simply count the number
 * of active ClassLoaders.
 * @param dest A Vector to hold the class loaders. If this is not null it will be cleared before
 * the classes are placed in it. If it is null then this just counts the number of active ClassLoaders.
 * @return The number of active class loaders.
 */
//===================================================================
public synchronized static int getClassLoaders(Vector dest)
//===================================================================
{
	return getClassLoaders(dest,false);
}
/**
 * Get the current set of application ClassLoaders into a Vector, or simply count the number
 * of active ClassLoaders.
 * @param dest A Vector to hold the class loaders. If this is not null it will be cleared before
 * the classes are placed in it. If it is null then this just counts the number of active ClassLoaders.
 * @param quick if this is true then the references are moved from the WeakSet directly
 * to the destination vector without editing out the possible null references. If
 * quick is false then the Vector will contain only non-null references.
 * @return The number of active class loaders.
 */
//===================================================================
public synchronized static int getClassLoaders(Vector dest, boolean quick)
//===================================================================
{
	if (dest != null) dest.clear();
	if (classLoaders == null) return 0;
 	if (classLoaders.isEmpty()) {
		classLoaders = null;
		return 0;
	}
 	//
 	if (quick) return classLoaders.getRefs(dest == null ? new Vector() : dest);
 	//
 	Vector v = new Vector();
	int did = classLoaders.getRefs(v);
	int num = 0;
	for (int i = 0; i<did; i++){
		if (v.get(i) == null) continue;
		num++;
		if (dest != null) dest.addElement(v.get(i));
	}
	return num;
}
//===================================================================
public mClassLoader()
//===================================================================
{
	addClassLoader(this);
}
/**
* This method will check for system classes and cached classes before calling
* findClassBytes().
**/
//-------------------------------------------------------------------
protected Class loadClass(String className,boolean resolve) throws ClassNotFoundException
//-------------------------------------------------------------------
{
	try{
		return Type.forName(className);
	}catch(ClassNotFoundException e){
		
	}
	Class c = findCachedClass(className);
	if (c == null){
		ByteArray data = findClassBytes(className);
		c = data == null ? null : defineClass(className,data.data,0,data.length);
	}else{
		//eve.sys.Vm.debug("Found in cache: "+className);
	}
	if (c == null) throw new ClassNotFoundException(className);
	if (resolve) resolveClass(c);
	cacheClass(className,c);
	return c;
}
/**
* A cache of already loaded classes.
**/
//-------------------------------------------------------------------
protected Hashtable loaded = new Hashtable();
//-------------------------------------------------------------------
/**
* Override this to provide class caching. By default a simple hashtable is used.
**/
//-------------------------------------------------------------------
protected Class findCachedClass(String className) {return (Class)loaded.get(className);}
//-------------------------------------------------------------------
/**
* Override this to provide class caching. By default a simple hashtable is used.
**/
//-------------------------------------------------------------------
protected void cacheClass(String className,Class aClass){loaded.put(className,aClass);}
//-------------------------------------------------------------------
/**
* You can override this method OR you can override getInputStreamFor(String className) - which is called
* by this method. It should attempt to locate the class
* bytes and put it in a returned ByteArray object. If it could not it should return null
* or throw a RuntimeException.
**/
//-------------------------------------------------------------------
protected ByteArray findClassBytes(String className)
//-------------------------------------------------------------------
{
	InputStream is = getInputStreamFor(className);
	if (is == null) return null;
	try{
		ByteArray ret = StreamUtils.readAllBytes(null,is,new ByteArray());
		is.close();
		return ret;
	}catch(IOException e){
		return null;
	}
}
/**
* This is called by findClassBytes by default and you can override this if necessary instead of overriding
* findClassBytes(). By default it will convert all '.' to '/' and then append '.class' to the end of it.
* It will then call openResource() with the converted class name. Therefore, overriding openResource() will
* take care of loading classes AND class dependant resources.
**/
//-------------------------------------------------------------------
protected InputStream getInputStreamFor(String className)
//-------------------------------------------------------------------
{
	className = className.replace('.','/')+".class";
	return openResource(className);
}
/**
* This is used to get a resource which may be dependant on the how the class
* was loaded.
**/
//===================================================================
public InputStream openResource(String resourceName)
//===================================================================
{
	if (sourceDirectory != null)
		try{
			return sourceDirectory.getChild(resourceName).toReadableStream();
		}catch(Exception e){
			return null;
		}
	if (fileStore != null)
		try{
			FileStoreEntry fe = fileStore.findEntry(resourceName);
			if (fe == null) {
				return null;
			}
			return fileStore.openInputStream(fe);
		}catch(Exception e){
			return null;
		}
	return null;
}
/**
 */
public static InputStream loadResourceFromAll(String resourceName)
{
	Vector v = new Vector();
	getClassLoaders(v,true);
	int s = v.size();
	for (int i = 0; i<s; i++){
		ClassLoader cl = (ClassLoader)v.get(i);
		if (!(cl instanceof mClassLoader)) continue;
		InputStream in = ((mClassLoader)cl).openResource(resourceName);
		if (in != null) return in;
	}
	return null;
}

/**
* This is a utility method used to load a Class by another Class which itself may have
* been loaded via a ClassLoader. It first tries to load the class via the default
* system class loader using Type.forName(). If this fails it checks the ClassLoader
* for the requestor class (if the requestor is not null) and requests that ClassLoader 
* to load the class. If that fails as well it will return null.
**/
/*
//===================================================================
public static Class getClass(String className,Class requestor)
//===================================================================
{
	try{
		return Type.forName(className);
	}catch(ClassNotFoundException e){
	}
	if (requestor == null) return null;
	ClassLoader cl = requestor.getClassLoader();
	if (cl == null) return null;
	try{
		return cl.loadClass(className);
	}catch(Exception ex){
		return null;
	}
}
*/
/**
* This tries to get a new instance of a class. First it calls mClassLoader.getClass()
* to get a Class object for the className. If that is successful it attempts to create
* a new instance of the class by calling the default constructor. If this is is successful
* it will return the newly created object.
**/
/*
//===================================================================
public static Object newInstance(String className,Class requestor)
//===================================================================
{
	Class c = getClass(className,requestor);
	if (c == null) return null;
	return Reflection.newInstance(c);
}
*/
/**
* This tries to open a resource via the class loader of the requestor. If this fails it will
* then try to open the resource normally. This actually calls ewe.sys.Vm.openResource().
**/
/*
//===================================================================
public static ewe.io.Stream openResource(String resourceName,Class requestor)
//===================================================================
{
	return ewe.sys.Vm.openResource(requestor,resourceName);	
}
*/
//##################################################################
}
//##################################################################


